import os
import shutil
import subprocess
import tempfile
from testtools import TestCase
from mock import (
    MagicMock,
    patch,
)
from charmhelpers.core.host import chdir

from urllib.parse import urlparse


try:
    from charmhelpers.fetch import (
        giturl,
        UnhandledSource,
    )
except ImportError:
    giturl = None
    UnhandledSource = None


class GitUrlFetchHandlerTest(TestCase):
    def setUp(self):
        super(GitUrlFetchHandlerTest, self).setUp()
        self.valid_urls = (
            "http://example.com/git-branch",
            "https://example.com/git-branch",
            "git://example.com/git-branch",
        )
        self.invalid_urls = (
            "file://example.com/foo.tar.bz2",
            "abc:example",
            "garbage",
        )
        self.fh = giturl.GitUrlFetchHandler()

    def test_handles_git_urls(self):
        for url in self.valid_urls:
            result = self.fh.can_handle(url)
            self.assertEqual(result, True, url)
        for url in self.invalid_urls:
            result = self.fh.can_handle(url)
            self.assertNotEqual(result, True, url)

    @patch.object(giturl, 'check_output')
    def test_clone(self, check_output):
        dest_path = "/destination/path"
        branch = "master"
        for url in self.valid_urls:
            self.fh.remote_branch = MagicMock()
            self.fh.load_plugins = MagicMock()
            self.fh.clone(url, dest_path, branch, None)

            check_output.assert_called_with(
                ['git', 'clone', url, dest_path, '--branch', branch], stderr=-2)

        for url in self.invalid_urls:
            with patch.dict('os.environ', {'CHARM_DIR': 'foo'}):
                self.assertRaises(UnhandledSource, self.fh.clone, url,
                                  dest_path, None,
                                  branch)

    def test_clone_functional(self):
        src = None
        dst = None
        try:
            src = tempfile.mkdtemp()
            with chdir(src):
                subprocess.check_output(['git', 'init'])
                subprocess.check_output(['git', 'config', 'user.name', 'Joe'])
                subprocess.check_output(
                    ['git', 'config', 'user.email', 'joe@test.com'])
                subprocess.check_output(['touch', 'foo'])
                subprocess.check_output(['git', 'add', 'foo'])
                subprocess.check_output(['git', 'commit', '-m', 'test'])
            dst = tempfile.mkdtemp()
            os.rmdir(dst)
            self.fh.clone(src, dst)
            assert os.path.exists(os.path.join(dst, '.git'))
            self.fh.clone(src, dst)  # idempotent
            assert os.path.exists(os.path.join(dst, '.git'))
        finally:
            if src:
                shutil.rmtree(src, ignore_errors=True)
            if dst:
                shutil.rmtree(dst, ignore_errors=True)

    def test_installs(self):
        self.fh.clone = MagicMock()

        for url in self.valid_urls:
            branch_name = urlparse(url).path.strip("/").split("/")[-1]
            dest = os.path.join('foo', 'fetched',
                                os.path.basename(branch_name))
            with patch.dict('os.environ', {'CHARM_DIR': 'foo'}):
                where = self.fh.install(url)
            self.assertEqual(where, dest)

    def test_installs_specified_dest(self):
        self.fh.clone = MagicMock()

        for url in self.valid_urls:
            branch_name = urlparse(url).path.strip("/").split("/")[-1]
            dest_repo = os.path.join('/tmp/git/',
                                     os.path.basename(branch_name))
            with patch.dict('os.environ', {'CHARM_DIR': 'foo'}):
                where = self.fh.install(url, dest="/tmp/git")
            self.assertEqual(where, dest_repo)



# Copyright 2014-2015 Canonical Limited.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#  http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os

from charmhelpers.core import host
from charmhelpers.core import hookenv


def render(source, target, context, owner='root', group='root',
           perms=0o444, templates_dir=None, encoding='UTF-8',
           template_loader=None, config_template=None):
    """
    Render a template.

    The `source` path, if not absolute, is relative to the `templates_dir`.

    The `target` path should be absolute.  It can also be `None`, in which
    case no file will be written.

    The context should be a dict containing the values to be replaced in the
    template.

    config_template may be provided to render from a provided template instead
    of loading from a file.

    The `owner`, `group`, and `perms` options will be passed to `write_file`.

    If omitted, `templates_dir` defaults to the `templates` folder in the charm.

    The rendered template will be written to the file as well as being returned
    as a string.

    Note: Using this requires python3-jinja2; if it is not installed, calling
    this will attempt to use charmhelpers.fetch.apt_install to install it.
    """
    try:
        from jinja2 import FileSystemLoader, Environment, exceptions
    except ImportError:
        try:
            from charmhelpers.fetch import apt_install
        except ImportError:
            hookenv.log('Could not import jinja2, and could not import '
                        'charmhelpers.fetch to install it',
                        level=hookenv.ERROR)
            raise
        apt_install('python3-jinja2', fatal=True)
        from jinja2 import FileSystemLoader, Environment, exceptions

    if template_loader:
        template_env = Environment(loader=template_loader)
    else:
        if templates_dir is None:
            templates_dir = os.path.join(hookenv.charm_dir(), 'templates')
        template_env = Environment(loader=FileSystemLoader(templates_dir))

    # load from a string if provided explicitly
    if config_template is not None:
        template = template_env.from_string(config_template)
    else:
        try:
            source = source
            template = template_env.get_template(source)
        except exceptions.TemplateNotFound as e:
            hookenv.log('Could not load template %s from %s.' %
                        (source, templates_dir),
                        level=hookenv.ERROR)
            raise e
    content = template.render(context)
    if target is not None:
        target_dir = os.path.dirname(target)
        if not os.path.exists(target_dir):
            # This is a terrible default directory permission, as the file
            # or its siblings will often contain secrets.
            host.mkdir(os.path.dirname(target), owner, group, perms=0o755)
        host.write_file(target, content.encode(encoding), owner, group, perms)
    return content



# Copyright 2016 Canonical Limited.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#  http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
import platform
import re
import subprocess

from charmhelpers.core.hookenv import (
    log,
    INFO,
    WARNING,
)
from charmhelpers.contrib.hardening import utils
from charmhelpers.contrib.hardening.audits.file import (
    FilePermissionAudit,
    TemplatedFile,
)
from charmhelpers.contrib.hardening.host import TEMPLATES_DIR


SYSCTL_DEFAULTS = """net.ipv4.ip_forward=%(net_ipv4_ip_forward)s
net.ipv6.conf.all.forwarding=%(net_ipv6_conf_all_forwarding)s
net.ipv4.conf.all.rp_filter=1
net.ipv4.conf.default.rp_filter=1
net.ipv4.icmp_echo_ignore_broadcasts=1
net.ipv4.icmp_ignore_bogus_error_responses=1
net.ipv4.icmp_ratelimit=100
net.ipv4.icmp_ratemask=88089
net.ipv6.conf.all.disable_ipv6=%(net_ipv6_conf_all_disable_ipv6)s
net.ipv4.tcp_timestamps=%(net_ipv4_tcp_timestamps)s
net.ipv4.conf.all.arp_ignore=%(net_ipv4_conf_all_arp_ignore)s
net.ipv4.conf.all.arp_announce=%(net_ipv4_conf_all_arp_announce)s
net.ipv4.tcp_rfc1337=1
net.ipv4.tcp_syncookies=1
net.ipv4.conf.all.shared_media=1
net.ipv4.conf.default.shared_media=1
net.ipv4.conf.all.accept_source_route=0
net.ipv4.conf.default.accept_source_route=0
net.ipv4.conf.all.accept_redirects=0
net.ipv4.conf.default.accept_redirects=0
net.ipv6.conf.all.accept_redirects=0
net.ipv6.conf.default.accept_redirects=0
net.ipv4.conf.all.secure_redirects=0
net.ipv4.conf.default.secure_redirects=0
net.ipv4.conf.all.send_redirects=0
net.ipv4.conf.default.send_redirects=0
net.ipv4.conf.all.log_martians=0
net.ipv6.conf.default.router_solicitations=0
net.ipv6.conf.default.accept_ra_rtr_pref=0
net.ipv6.conf.default.accept_ra_pinfo=0
net.ipv6.conf.default.accept_ra_defrtr=0
net.ipv6.conf.default.autoconf=0
net.ipv6.conf.default.dad_transmits=0
net.ipv6.conf.default.max_addresses=1
net.ipv6.conf.all.accept_ra=0
net.ipv6.conf.default.accept_ra=0
kernel.modules_disabled=%(kernel_modules_disabled)s
kernel.sysrq=%(kernel_sysrq)s
fs.suid_dumpable=%(fs_suid_dumpable)s
kernel.randomize_va_space=2
"""


def get_audits():
    """Get OS hardening sysctl audits.

    :returns:  dictionary of audits
    """
    audits = []
    settings = utils.get_settings('os')

    # Apply the sysctl settings which are configured to be applied.
    audits.append(SysctlConf())
    # Make sure that only root has access to the sysctl.conf file, and
    # that it is read-only.
    audits.append(FilePermissionAudit('/etc/sysctl.conf',
                                      user='root',
                                      group='root', mode=0o0440))
    # If module loading is not enabled, then ensure that the modules
    # file has the appropriate permissions and rebuild the initramfs
    if not settings['security']['kernel_enable_module_loading']:
        audits.append(ModulesTemplate())

    return audits


class ModulesContext(object):

    def __call__(self):
        settings = utils.get_settings('os')
        with open('/proc/cpuinfo', 'r') as fd:
            cpuinfo = fd.readlines()

        for line in cpuinfo:
            match = re.search(r"^vendor_id\s+:\s+(.+)", line)
            if match:
                vendor = match.group(1)

        if vendor == "GenuineIntel":
            vendor = "intel"
        elif vendor == "AuthenticAMD":
            vendor = "amd"

        ctxt = {'arch': platform.processor(),
                'cpuVendor': vendor,
                'desktop_enable': settings['general']['desktop_enable']}

        return ctxt


class ModulesTemplate(object):

    def __init__(self):
        super(ModulesTemplate, self).__init__('/etc/initramfs-tools/modules',
                                              ModulesContext(),
                                              templates_dir=TEMPLATES_DIR,
                                              user='root', group='root',
                                              mode=0o0440)

    def post_write(self):
        subprocess.check_call(['update-initramfs', '-u'])


class SysCtlHardeningContext(object):
    def __call__(self):
        settings = utils.get_settings('os')
        ctxt = {'sysctl': {}}

        log("Applying sysctl settings", level=INFO)
        extras = {'net_ipv4_ip_forward': 0,
                  'net_ipv6_conf_all_forwarding': 0,
                  'net_ipv6_conf_all_disable_ipv6': 1,
                  'net_ipv4_tcp_timestamps': 0,
                  'net_ipv4_conf_all_arp_ignore': 0,
                  'net_ipv4_conf_all_arp_announce': 0,
                  'kernel_sysrq': 0,
                  'fs_suid_dumpable': 0,
                  'kernel_modules_disabled': 1}

        if settings['sysctl']['ipv6_enable']:
            extras['net_ipv6_conf_all_disable_ipv6'] = 0

        if settings['sysctl']['forwarding']:
            extras['net_ipv4_ip_forward'] = 1
            extras['net_ipv6_conf_all_forwarding'] = 1

        if settings['sysctl']['arp_restricted']:
            extras['net_ipv4_conf_all_arp_ignore'] = 1
            extras['net_ipv4_conf_all_arp_announce'] = 2

        if settings['security']['kernel_enable_module_loading']:
            extras['kernel_modules_disabled'] = 0

        if settings['sysctl']['kernel_enable_sysrq']:
            sysrq_val = settings['sysctl']['kernel_secure_sysrq']
            extras['kernel_sysrq'] = sysrq_val

        if settings['security']['kernel_enable_core_dump']:
            extras['fs_suid_dumpable'] = 1

        settings.update(extras)
        for d in (SYSCTL_DEFAULTS % settings).split():
            d = d.strip().partition('=')
            key = d[0].strip()
            path = os.path.join('/proc/sys', key.replace('.', '/'))
            if not os.path.exists(path):
                log("Skipping '%s' since '%s' does not exist" % (key, path),
                    level=WARNING)
                continue

            ctxt['sysctl'][key] = d[2] or None

        return {
            'sysctl_settings': [(k, v) for k, v in ctxt['sysctl'].items()]
        }


class SysctlConf(TemplatedFile):
    """An audit check for sysctl settings."""
    def __init__(self):
        self.conffile = '/etc/sysctl.d/99-juju-hardening.conf'
        super(SysctlConf, self).__init__(self.conffile,
                                         SysCtlHardeningContext(),
                                         template_dir=TEMPLATES_DIR,
                                         user='root', group='root',
                                         mode=0o0440)

    def post_write(self):
        try:
            subprocess.check_call(['sysctl', '-p', self.conffile])
        except subprocess.CalledProcessError as e:
            # NOTE: on some systems if sysctl cannot apply all settings it
            #       will return non-zero as well.
            log("sysctl command returned an error (maybe some "
                "keys could not be set) - %s" % (e),
                level=WARNING)



import contextlib
import copy
import datetime
import json
import tempfile
import shutil
import yaml

from mock import patch, call

import charmhelpers.contrib.openstack.deferred_events as deferred_events
import tests.utils


class TestDB(object):
    '''Test KV store for unitdata testing'''
    def __init__(self):
        self.data = {}
        self.flushed = False

    def get(self, key, default=None):
        result = self.data.get(key, default)
        if not result:
            return default
        return json.loads(result)

    def set(self, key, value):
        self.data[key] = json.dumps(value)
        return value

    def flush(self):
        self.flushed = True


class TestHookData(object):

    def __init__(self, kv):
        self.kv = kv

    @contextlib.contextmanager
    def __call__(self):
        yield self.kv, True, True


class DeferredCharmServiceEventsTestCase(tests.utils.BaseTestCase):

    def setUp(self):
        super(DeferredCharmServiceEventsTestCase, self).setUp()
        self.tmp_dir = tempfile.mkdtemp()
        self.addCleanup(lambda: shutil.rmtree(self.tmp_dir))
        self.patch_object(deferred_events.hookenv, 'service_name')
        self.service_name.return_value = 'myapp'
        self.patch_object(deferred_events.unitdata, 'HookData')
        self.db = TestDB()
        self.HookData.return_value = TestHookData(self.db)
        self.exp_event_a = deferred_events.ServiceEvent(
            timestamp=123,
            service='svcA',
            reason='ReasonA',
            action='restart',
            policy_requestor_name='myapp',
            policy_requestor_type='charm')
        self.exp_event_b = deferred_events.ServiceEvent(
            timestamp=223,
            service='svcB',
            reason='ReasonB',
            action='restart')
        self.exp_event_c = deferred_events.ServiceEvent(
            timestamp=323,
            service='svcB',
            reason='ReasonB',
            action='restart')
        self.base_expect_events = [
            self.exp_event_a,
            self.exp_event_b,
            self.exp_event_c]
        self.event_file_pair = []
        for index, event in enumerate(self.base_expect_events):
            event_file = '{}/{}.deferred'.format('/tmpdir', str(index))
            self.event_file_pair.append((
                event_file,
                event))

    def test_matching_request_event(self):
        self.assertTrue(
            self.exp_event_b.matching_request(
                self.exp_event_c))
        self.assertFalse(
            self.exp_event_a.matching_request(
                self.exp_event_b))

    @patch.object(deferred_events.glob, "glob")
    def test_deferred_events_files(self, glob):
        defer_files = [
            '/var/lib/policy-rc.d/charm-myapp/1612346300.deferred',
            '/var/lib/policy-rc.d/charm-myapp/1612346322.deferred',
            '/var/lib/policy-rc.d/charm-myapp/1612346360.deferred']

        glob.return_value = defer_files
        self.assertEqual(
            deferred_events.deferred_events_files(),
            defer_files)

    def test_read_event_file(self):
        with tempfile.NamedTemporaryFile('w') as ftmp:
            yaml.dump(vars(self.exp_event_a), ftmp)
            ftmp.flush()
            self.assertEqual(
                deferred_events.read_event_file(ftmp.name),
                self.exp_event_a)

    @patch.object(deferred_events, "deferred_events_files")
    def test_deferred_events(self, deferred_events_files):
        event_files = []
        expect = []
        for index, event in enumerate(self.base_expect_events):
            event_file = '{}/{}.deferred'.format(self.tmp_dir, str(index))
            with open(event_file, 'w') as f:
                yaml.dump(vars(event), f)
            event_files.append(event_file)
            expect.append((
                event_file,
                event))
        deferred_events_files.return_value = event_files
        self.assertEqual(
            deferred_events.deferred_events(),
            expect)

    @patch.object(deferred_events, "deferred_events")
    def test_duplicate_event_files(self, _deferred_events):
        _deferred_events.return_value = self.event_file_pair
        self.assertEqual(
            deferred_events.duplicate_event_files(self.exp_event_b),
            ['/tmpdir/1.deferred', '/tmpdir/2.deferred'])
        self.assertEqual(
            deferred_events.duplicate_event_files(deferred_events.ServiceEvent(
                timestamp=223,
                service='svcX',
                reason='ReasonX',
                action='restart')),
            [])

    @patch.object(deferred_events.uuid, "uuid1")
    def test_get_event_record_file(self, uuid1):
        uuid1.return_value = '89eb8258'
        self.assertEqual(
            deferred_events.get_event_record_file(
                'charm',
                'neutron-ovs'),
            '/var/lib/policy-rc.d/charm-neutron-ovs-89eb8258.deferred')

    @patch.object(deferred_events, "get_event_record_file")
    @patch.object(deferred_events, "duplicate_event_files")
    @patch.object(deferred_events, "init_policy_log_dir")
    def test_save_event(self, init_policy_log_dir, duplicate_event_files, get_event_record_file):
        duplicate_event_files.return_value = []
        test_file = '{}/test_save_event.yaml'.format(self.tmp_dir)
        get_event_record_file.return_value = test_file
        deferred_events.save_event(self.exp_event_a)
        with open(test_file, 'r') as f:
            contents = yaml.safe_load(f)
        self.assertEqual(contents, vars(self.exp_event_a))

    @patch.object(deferred_events.os, "remove")
    @patch.object(deferred_events, "read_event_file")
    @patch.object(deferred_events, "deferred_events_files")
    def test_clear_deferred_events(self, deferred_events_files, read_event_file,
                                   remove):
        deferred_events_files.return_value = ['/tmp/file1']
        read_event_file.return_value = self.exp_event_a
        deferred_events.clear_deferred_events('svcB', 'restart')
        self.assertFalse(remove.called)
        deferred_events.clear_deferred_events('svcA', 'restart')
        remove.assert_called_once_with('/tmp/file1')

    @patch.object(deferred_events.os, "mkdir")
    @patch.object(deferred_events.os.path, "exists")
    def test_init_policy_log_dir(self, exists, mkdir):
        exists.return_value = True
        deferred_events.init_policy_log_dir()
        self.assertFalse(mkdir.called)
        exists.return_value = False
        deferred_events.init_policy_log_dir()
        mkdir.assert_called_once_with('/var/lib/policy-rc.d')

    @patch.object(deferred_events, "deferred_events")
    def test_get_deferred_events(self, _deferred_events):
        _deferred_events.return_value = self.event_file_pair
        self.assertEqual(
            deferred_events.get_deferred_events(),
            self.base_expect_events)

    @patch.object(deferred_events, "get_deferred_events")
    def test_get_deferred_restarts(self, get_deferred_events):
        test_events = copy.deepcopy(self.base_expect_events)
        test_events.append(
            deferred_events.ServiceEvent(
                timestamp=523,
                service='svcD',
                reason='StopReasonD',
                action='stop'))
        get_deferred_events.return_value = test_events
        self.assertEqual(
            deferred_events.get_deferred_restarts(),
            self.base_expect_events)

    @patch.object(deferred_events, 'clear_deferred_events')
    def test_clear_deferred_restarts(self, clear_deferred_events):
        deferred_events.clear_deferred_restarts(['svcA', 'svcB'])
        clear_deferred_events.assert_Called_once_with(
            ['svcA', 'svcB'],
            'restart')

    @patch.object(deferred_events, 'clear_deferred_restarts')
    def test_process_svc_restart(self, clear_deferred_restarts):
        deferred_events.process_svc_restart('svcA')
        clear_deferred_restarts.assert_called_once_with(
            ['svcA'])

    @patch.object(deferred_events.hookenv, 'config')
    def test_is_restart_permitted(self, config):
        config.return_value = None
        self.assertTrue(deferred_events.is_restart_permitted())
        config.return_value = True
        self.assertTrue(deferred_events.is_restart_permitted())
        config.return_value = False
        self.assertFalse(deferred_events.is_restart_permitted())

    @patch.object(deferred_events.time, 'time')
    @patch.object(deferred_events, 'save_event')
    @patch.object(deferred_events, 'is_restart_permitted')
    def test_check_and_record_restart_request(self, is_restart_permitted,
                                              save_event, time):
        time.return_value = 123
        is_restart_permitted.return_value = False
        deferred_events.check_and_record_restart_request(
            'svcA',
            ['/tmp/test1.conf', '/tmp/test2.conf'])
        save_event.assert_called_once_with(deferred_events.ServiceEvent(
            timestamp=123,
            service='svcA',
            reason='File(s) changed: /tmp/test1.conf, /tmp/test2.conf',
            action='restart'))

    @patch.object(deferred_events.time, 'time')
    @patch.object(deferred_events, 'save_event')
    @patch.object(deferred_events.host, 'service_restart')
    @patch.object(deferred_events, 'is_restart_permitted')
    def test_deferrable_svc_restart(self, is_restart_permitted,
                                    service_restart, save_event, time):
        time.return_value = 123
        is_restart_permitted.return_value = True
        deferred_events.deferrable_svc_restart('svcA', reason='ReasonA')
        service_restart.assert_called_once_with('svcA')
        service_restart.reset_mock()
        is_restart_permitted.return_value = False
        deferred_events.deferrable_svc_restart('svcA', reason='ReasonA')
        self.assertFalse(service_restart.called)
        save_event.assert_called_once_with(deferred_events.ServiceEvent(
            timestamp=123,
            service='svcA',
            reason='ReasonA',
            action='restart'))

    @patch.object(deferred_events.policy_rcd, 'add_policy_block')
    @patch.object(deferred_events.policy_rcd, 'remove_policy_file')
    @patch.object(deferred_events, 'is_restart_permitted')
    @patch.object(deferred_events.policy_rcd, 'install_policy_rcd')
    def test_configure_deferred_restarts(self, install_policy_rcd,
                                         is_restart_permitted,
                                         remove_policy_file, add_policy_block):
        is_restart_permitted.return_value = True
        deferred_events.configure_deferred_restarts(['svcA', 'svcB'])
        remove_policy_file.assert_called_once_with()
        install_policy_rcd.assert_called_once_with()

        remove_policy_file.reset_mock()
        install_policy_rcd.reset_mock()
        is_restart_permitted.return_value = False
        deferred_events.configure_deferred_restarts(['svcA', 'svcB'])
        self.assertFalse(remove_policy_file.called)
        install_policy_rcd.assert_called_once_with()
        add_policy_block.assert_has_calls([
            call('svcA', ['stop', 'restart', 'try-restart']),
            call('svcB', ['stop', 'restart', 'try-restart'])])

    @patch.object(deferred_events.subprocess, 'check_output')
    def test_get_service_start_time(self, check_output):
        check_output.return_value = (
            b'ActiveEnterTimestamp=Tue 2021-02-02 13:19:55 UTC')
        expect = datetime.datetime.strptime(
            'Tue 2021-02-02 13:19:55 UTC',
            '%a %Y-%m-%d %H:%M:%S %Z')
        self.assertEqual(
            deferred_events.get_service_start_time('svcA'),
            expect)
        check_output.assert_called_once_with(
            ['systemctl', 'show', 'svcA', '--property=ActiveEnterTimestamp'])

    @patch.object(deferred_events, 'get_deferred_restarts')
    @patch.object(deferred_events, 'clear_deferred_restarts')
    @patch.object(deferred_events.hookenv, 'log')
    @patch.object(deferred_events, 'get_service_start_time')
    def test_check_restart_timestamps(self, get_service_start_time, log,
                                      clear_deferred_restarts,
                                      get_deferred_restarts):
        deferred_restarts = [
            # 'Tue 2021-02-02 10:19:55 UTC'
            deferred_events.ServiceEvent(
                timestamp=1612261195.0,
                service='svcA',
                reason='ReasonA',
                action='restart')]
        get_deferred_restarts.return_value = deferred_restarts
        get_service_start_time.return_value = datetime.datetime.strptime(
            'Tue 2021-02-02 13:19:55 UTC',
            '%a %Y-%m-%d %H:%M:%S %Z')
        deferred_events.check_restart_timestamps()
        clear_deferred_restarts.assert_called_once_with(['svcA'])

        clear_deferred_restarts.reset_mock()
        get_service_start_time.return_value = datetime.datetime.strptime(
            'Tue 2021-02-02 10:10:55 UTC',
            '%a %Y-%m-%d %H:%M:%S %Z')
        deferred_events.check_restart_timestamps()
        self.assertFalse(clear_deferred_restarts.called)
        log.assert_called_once_with(
            ('Restart still required, svcA was started at 2021-02-02 10:10:55,'
             ' restart was requested after that at {}'.format(
                 datetime.datetime.fromtimestamp(1612261195.0))),
            level='DEBUG')

    def test_set_deferred_hook(self):
        deferred_events.set_deferred_hook('config-changed')
        self.assertEqual(self.db.get('deferred-hooks'), ['config-changed'])
        deferred_events.set_deferred_hook('leader-settings-changed')
        self.assertEqual(
            self.db.get('deferred-hooks'),
            ['config-changed', 'leader-settings-changed'])

    def test_get_deferred_hook(self):
        deferred_events.set_deferred_hook('config-changed')
        self.assertEqual(
            deferred_events.get_deferred_hooks(),
            ['config-changed'])

    def test_clear_deferred_hooks(self):
        deferred_events.set_deferred_hook('config-changed')
        deferred_events.set_deferred_hook('leader-settings-changed')
        self.assertEqual(
            deferred_events.get_deferred_hooks(),
            ['config-changed', 'leader-settings-changed'])
        deferred_events.clear_deferred_hooks()
        self.assertEqual(
            deferred_events.get_deferred_hooks(),
            [])

    def test_clear_deferred_hook(self):
        deferred_events.set_deferred_hook('config-changed')
        deferred_events.set_deferred_hook('leader-settings-changed')
        self.assertEqual(
            deferred_events.get_deferred_hooks(),
            ['config-changed', 'leader-settings-changed'])
        deferred_events.clear_deferred_hook('leader-settings-changed')
        self.assertEqual(
            deferred_events.get_deferred_hooks(),
            ['config-changed'])



# Copyright 2014-2015 Canonical Limited.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#  http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# Easy file synchronization among peer units using ssh + unison.
#
# For the -joined, -changed, and -departed peer relations, add a call to
# ssh_authorized_peers() describing the peer relation and the desired
# user + group.  After all peer relations have settled, all hosts should
# be able to connect to on another via key auth'd ssh as the specified user.
#
# Other hooks are then free to synchronize files and directories using
# sync_to_peers().
#
# For a peer relation named 'cluster', for example:
#
# cluster-relation-joined:
# ...
# ssh_authorized_peers(peer_interface='cluster',
#                      user='juju_ssh', group='juju_ssh',
#                      ensure_local_user=True)
# ...
#
# cluster-relation-changed:
# ...
# ssh_authorized_peers(peer_interface='cluster',
#                      user='juju_ssh', group='juju_ssh',
#                      ensure_local_user=True)
# ...
#
# cluster-relation-departed:
# ...
# ssh_authorized_peers(peer_interface='cluster',
#                      user='juju_ssh', group='juju_ssh',
#                      ensure_local_user=True)
# ...
#
# Hooks are now free to sync files as easily as:
#
# files = ['/etc/fstab', '/etc/apt.conf.d/']
# sync_to_peers(peer_interface='cluster',
#                user='juju_ssh, paths=[files])
#
# It is assumed the charm itself has setup permissions on each unit
# such that 'juju_ssh' has read + write permissions.  Also assumed
# that the calling charm takes care of leader delegation.
#
# Additionally files can be synchronized only to an specific unit:
# sync_to_peer(slave_address, user='juju_ssh',
#              paths=[files], verbose=False)

import os
import pwd

from copy import copy
from subprocess import check_call, check_output

from charmhelpers.core.host import (
    adduser,
    add_user_to_group,
    pwgen,
    remove_password_expiry,
)

from charmhelpers.core.hookenv import (
    log,
    hook_name,
    relation_ids,
    related_units,
    relation_set,
    relation_get,
    unit_private_ip,
    INFO,
    ERROR,
)

BASE_CMD = ['unison', '-auto', '-batch=true', '-confirmbigdel=false',
            '-fastcheck=true', '-group=false', '-owner=false',
            '-prefer=newer', '-times=true']


def get_homedir(user):
    try:
        user = pwd.getpwnam(user)
        return user.pw_dir
    except KeyError:
        log('Could not get homedir for user %s: user exists?' % (user), ERROR)
        raise Exception


def create_private_key(user, priv_key_path, key_type='rsa'):
    types_bits = {
        'rsa': '2048',
        'ecdsa': '521',
    }
    if key_type not in types_bits:
        log('Unknown ssh key type {}, using rsa'.format(key_type), ERROR)
        key_type = 'rsa'
    if not os.path.isfile(priv_key_path):
        log('Generating new SSH key for user %s.' % user)
        cmd = ['ssh-keygen', '-q', '-N', '', '-t', key_type,
               '-b', types_bits[key_type], '-f', priv_key_path]
        check_call(cmd)
    else:
        log('SSH key already exists at %s.' % priv_key_path)
    os.chown(priv_key_path, pwd.getpwnam(user).pw_uid, -1)
    os.chmod(priv_key_path, 0o600)


def create_public_key(user, priv_key_path, pub_key_path):
    if not os.path.isfile(pub_key_path):
        log('Generating missing ssh public key @ %s.' % pub_key_path)
        cmd = ['ssh-keygen', '-y', '-f', priv_key_path]
        p = check_output(cmd).strip()
        with open(pub_key_path, 'wb') as out:
            out.write(p)
    os.chown(pub_key_path, pwd.getpwnam(user).pw_uid, -1)


def get_keypair(user):
    home_dir = get_homedir(user)
    ssh_dir = os.path.join(home_dir, '.ssh')
    priv_key = os.path.join(ssh_dir, 'id_rsa')
    pub_key = '%s.pub' % priv_key

    if not os.path.isdir(ssh_dir):
        os.mkdir(ssh_dir)
        check_call(['chown', '-R', user, ssh_dir])

    create_private_key(user, priv_key)
    create_public_key(user, priv_key, pub_key)

    with open(priv_key, 'r') as p:
        _priv = p.read().strip()

    with open(pub_key, 'r') as p:
        _pub = p.read().strip()

    return (_priv, _pub)


def write_authorized_keys(user, keys):
    home_dir = get_homedir(user)
    ssh_dir = os.path.join(home_dir, '.ssh')
    auth_keys = os.path.join(ssh_dir, 'authorized_keys')
    log('Syncing authorized_keys @ %s.' % auth_keys)
    with open(auth_keys, 'w') as out:
        for k in keys:
            out.write('%s\n' % k)
    os.chown(auth_keys, pwd.getpwnam(user).pw_uid, -1)


def write_known_hosts(user, hosts):
    home_dir = get_homedir(user)
    ssh_dir = os.path.join(home_dir, '.ssh')
    known_hosts = os.path.join(ssh_dir, 'known_hosts')
    khosts = []
    for host in hosts:
        cmd = ['ssh-keyscan', host]
        remote_key = check_output(cmd, universal_newlines=True).strip()
        khosts.append(remote_key)
    log('Syncing known_hosts @ %s.' % known_hosts)
    with open(known_hosts, 'w') as out:
        for host in khosts:
            out.write('%s\n' % host)
    os.chown(known_hosts, pwd.getpwnam(user).pw_uid, -1)


def ensure_user(user, group=None):
    adduser(user, pwgen())
    if group:
        add_user_to_group(user, group)
    # Remove password expiry (Bug #1686085)
    remove_password_expiry(user)


def ssh_authorized_peers(peer_interface, user, group=None,
                         ensure_local_user=False):
    """
    Main setup function, should be called from both peer -changed and -joined
    hooks with the same parameters.
    """
    if ensure_local_user:
        ensure_user(user, group)
    priv_key, pub_key = get_keypair(user)
    hook = hook_name()
    if hook == '%s-relation-joined' % peer_interface:
        relation_set(ssh_pub_key=pub_key)
    elif hook == '%s-relation-changed' % peer_interface or \
            hook == '%s-relation-departed' % peer_interface:
        hosts = []
        keys = []

        for r_id in relation_ids(peer_interface):
            for unit in related_units(r_id):
                ssh_pub_key = relation_get('ssh_pub_key',
                                           rid=r_id,
                                           unit=unit)
                priv_addr = relation_get('private-address',
                                         rid=r_id,
                                         unit=unit)
                if ssh_pub_key:
                    keys.append(ssh_pub_key)
                    hosts.append(priv_addr)
                else:
                    log('ssh_authorized_peers(): ssh_pub_key '
                        'missing for unit %s, skipping.' % unit)
        write_authorized_keys(user, keys)
        write_known_hosts(user, hosts)
        authed_hosts = ':'.join(hosts)
        relation_set(ssh_authorized_hosts=authed_hosts)


def _run_as_user(user, gid=None):
    try:
        user = pwd.getpwnam(user)
    except KeyError:
        log('Invalid user: %s' % user)
        raise Exception
    uid = user.pw_uid
    gid = gid or user.pw_gid
    os.environ['HOME'] = user.pw_dir

    def _inner():
        os.setgid(gid)
        os.setuid(uid)
    return _inner


def run_as_user(user, cmd, gid=None):
    return check_output(cmd, preexec_fn=_run_as_user(user, gid), cwd='/')


def collect_authed_hosts(peer_interface):
    '''Iterate through the units on peer interface to find all that
    have the calling host in its authorized hosts list'''
    hosts = []
    for r_id in (relation_ids(peer_interface) or []):
        for unit in related_units(r_id):
            private_addr = relation_get('private-address',
                                        rid=r_id, unit=unit)
            authed_hosts = relation_get('ssh_authorized_hosts',
                                        rid=r_id, unit=unit)

            if not authed_hosts:
                log('Peer %s has not authorized *any* hosts yet, skipping.' %
                    (unit), level=INFO)
                continue

            if unit_private_ip() in authed_hosts.split(':'):
                hosts.append(private_addr)
            else:
                log('Peer %s has not authorized *this* host yet, skipping.' %
                    (unit), level=INFO)
    return hosts


def sync_path_to_host(path, host, user, verbose=False, cmd=None, gid=None,
                      fatal=False):
    """Sync path to an specific peer host

    Propagates exception if operation fails and fatal=True.
    """
    cmd = cmd or copy(BASE_CMD)
    if not verbose:
        cmd.append('-silent')

    # removing trailing slash from directory paths, unison
    # doesn't like these.
    if path.endswith('/'):
        path = path[:(len(path) - 1)]

    cmd = cmd + [path, 'ssh://%s@%s/%s' % (user, host, path)]

    try:
        log('Syncing local path %s to %s@%s:%s' % (path, user, host, path))
        run_as_user(user, cmd, gid)
    except Exception:
        log('Error syncing remote files')
        if fatal:
            raise


def sync_to_peer(host, user, paths=None, verbose=False, cmd=None, gid=None,
                 fatal=False):
    """Sync paths to an specific peer host

    Propagates exception if any operation fails and fatal=True.
    """
    if paths:
        for p in paths:
            sync_path_to_host(p, host, user, verbose, cmd, gid, fatal)


def sync_to_peers(peer_interface, user, paths=None, verbose=False, cmd=None,
                  gid=None, fatal=False):
    """Sync all hosts to an specific path

    The type of group is integer, it allows user has permissions to
    operate a directory have a different group id with the user id.

    Propagates exception if any operation fails and fatal=True.
    """
    if paths:
        for host in collect_authed_hosts(peer_interface):
            sync_to_peer(host, user, paths, verbose, cmd, gid, fatal)



import sys
import mock


sys.modules['yum'] = mock.MagicMock()
with mock.patch('charmhelpers.deprecate') as ch_deprecate:
    def mock_deprecate(warning, date=None, log=None):
        def mock_wrap(f):
            def wrapped_f(*args, **kwargs):
                return f(*args, **kwargs)
            return wrapped_f
        return mock_wrap
    ch_deprecate.side_effect = mock_deprecate
    import charmhelpers.contrib.openstack.utils as openstack  # noqa: F401



#!/usr/bin/env python
# coding: utf-8

# Copyright 2014-2015 Canonical Limited.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#  http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import sys

__author__ = "Jorge Niedbalski <jorge.niedbalski@canonical.com>"


def current_version():
    """Current system python version"""
    return sys.version_info


def current_version_string():
    """Current system python version as string major.minor.micro"""
    return "{0}.{1}.{2}".format(sys.version_info.major,
                                sys.version_info.minor,
                                sys.version_info.micro)



# -*- coding: utf-8 -*-
#
# Copyright 2015 Canonical Ltd.
#
# Authors:
#  Kapil Thangavelu <kapil.foss@gmail.com>
#
try:
    from StringIO import StringIO
except Exception:
    from io import StringIO

import os
import shutil
import tempfile
import unittest

from mock import patch

from charmhelpers.core.unitdata import Storage, HookData, kv


class HookDataTest(unittest.TestCase):

    def setUp(self):
        self.charm_dir = tempfile.mkdtemp()
        self.addCleanup(lambda: shutil.rmtree(self.charm_dir))
        self.change_environment(CHARM_DIR=self.charm_dir)

    def change_environment(self, **kw):
        original_env = dict(os.environ)

        @self.addCleanup
        def cleanup_env():
            os.environ.clear()
            os.environ.update(original_env)

        os.environ.update(kw)

    @patch('charmhelpers.core.hookenv.hook_name')
    @patch('charmhelpers.core.hookenv.execution_environment')
    @patch('charmhelpers.core.hookenv.charm_dir')
    def test_hook_data_records(self, cdir, ctx, name):
        name.return_value = 'config-changed'
        ctx.return_value = {
            'rels': {}, 'conf': {'a': 1}, 'env': {}, 'unit': 'someunit'}
        cdir.return_value = self.charm_dir
        with open(os.path.join(self.charm_dir, 'revision'), 'w') as fh:
            fh.write('1')
        hook_data = HookData()

        with hook_data():
            self.assertEqual(kv(), hook_data.kv)
            self.assertEqual(kv().get('charm_revisions'), ['1'])
            self.assertEqual(kv().get('unit'), 'someunit')
            self.assertEqual(list(hook_data.conf), ['a'])
            self.assertEqual(tuple(hook_data.conf.a), (None, 1))


class StorageTest(unittest.TestCase):

    def test_init_kv_multiple(self):
        with tempfile.NamedTemporaryFile() as fh:
            kv = Storage(fh.name)
            with kv.hook_scope('xyz'):
                kv.set('x', 1)
            kv.close()
            self.assertEqual(os.stat(fh.name).st_mode & 0o777, 0o600)

            kv = Storage(fh.name)
            with kv.hook_scope('abc'):
                self.assertEqual(kv.get('x'), 1)
            kv.close()

    def test_hook_scope(self):
        kv = Storage(':memory:')
        try:
            with kv.hook_scope('install') as rev:
                self.assertEqual(rev, 1)
                kv.set('a', 1)
                raise RuntimeError('x')
        except RuntimeError:
            self.assertEqual(kv.get('a'), None)

        with kv.hook_scope('config-changed') as rev:
            self.assertEqual(rev, 1)
            kv.set('a', 1)
        self.assertEqual(kv.get('a'), 1)

        kv.revision = None

        with kv.hook_scope('start') as rev:
            self.assertEqual(rev, 2)
            kv.set('a', False)
            kv.set('a', True)
        self.assertEqual(kv.get('a'), True)

        # History doesn't decode values by default
        history = [h[:-1] for h in kv.gethistory('a')]
        self.assertEqual(
            history,
            [(1, 'a', '1', 'config-changed'),
             (2, 'a', 'true', 'start')])

        history = [h[:-1] for h in kv.gethistory('a', deserialize=True)]
        self.assertEqual(
            history,
            [(1, 'a', 1, 'config-changed'),
             (2, 'a', True, 'start')])

    def test_delta_no_previous_and_history(self):
        kv = Storage(':memory:')
        with kv.hook_scope('install'):
            data = {'a': 0, 'c': False}
            delta = kv.delta(data, 'settings.')
            self.assertEqual(delta, {
                'a': (None, False), 'c': (None, False)})
            kv.update(data, 'settings.')

        with kv.hook_scope('config'):
            data = {'a': 1, 'c': True}
            delta = kv.delta(data, 'settings.')
            self.assertEqual(delta, {
                'a': (0, 1), 'c': (False, True)})
            kv.update(data, 'settings.')
        # strip the time
        history = [h[:-1] for h in kv.gethistory('settings.a')]
        self.assertEqual(
            history,
            [(1, 'settings.a', '0', 'install'),
             (2, 'settings.a', '1', 'config')])

    def test_unset(self):
        kv = Storage(':memory:')
        with kv.hook_scope('install'):
            kv.set('a', True)
        with kv.hook_scope('start'):
            kv.set('a', False)
        with kv.hook_scope('config-changed'):
            kv.unset('a')
        history = [h[:-1] for h in kv.gethistory('a')]

        self.assertEqual(history, [
            (1, 'a', 'true', 'install'),
            (2, 'a', 'false', 'start'),
            (3, 'a', '"DELETED"', "config-changed")])

    def test_flush_and_close_on_closed(self):
        kv = Storage(':memory:')
        kv.close()
        kv.flush(False)
        kv.close()

    def test_multi_value_set_skips(self):
        # pure coverage test
        kv = Storage(':memory:')
        kv.set('x', 1)
        self.assertEqual(kv.set('x', 1), 1)

    def test_debug(self):
        # pure coverage test...
        io = StringIO()
        kv = Storage(':memory:')
        kv.debug(io)

    def test_record(self):
        kv = Storage(':memory:')
        kv.set('config', {'x': 1, 'b': False})
        config = kv.get('config', record=True)
        self.assertEqual(config.b, False)
        self.assertEqual(config.x, 1)
        self.assertEqual(kv.set('config.x', 1), 1)
        try:
            config.z
        except AttributeError:
            pass
        else:
            self.fail('attribute error should fire on nonexistant')

    def test_delta(self):
        kv = Storage(':memory:')
        kv.update({'a': 1, 'b': 2.2}, prefix="x")
        delta = kv.delta({'a': 0, 'c': False}, prefix='x')
        self.assertEqual(
            delta,
            {'a': (1, 0), 'b': (2.2, None), 'c': (None, False)})
        self.assertEqual(delta.a.previous, 1)
        self.assertEqual(delta.a.current, 0)
        self.assertEqual(delta.c.previous, None)
        self.assertEqual(delta.a.current, False)

    def test_update(self):
        kv = Storage(':memory:')
        kv.update({'v_a': 1, 'v_b': 2.2})
        self.assertEqual(kv.getrange('v_'), {'v_a': 1, 'v_b': 2.2})

        kv.update({'a': False, 'b': True}, prefix='x_')
        self.assertEqual(
            kv.getrange('x_', True), {'a': False, 'b': True})

    def test_keyrange(self):
        kv = Storage(':memory:')
        kv.set('docker.net_mtu', 1)
        kv.set('docker.net_nack', True)
        kv.set('docker.net_type', 'vxlan')
        self.assertEqual(
            kv.getrange('docker'),
            {'docker.net_mtu': 1, 'docker.net_type': 'vxlan',
             'docker.net_nack': True})
        self.assertEqual(
            kv.getrange('docker.', True),
            {'net_mtu': 1, 'net_type': 'vxlan', 'net_nack': True})

    def test_get_set_unset(self):
        kv = Storage(':memory:')
        kv.hook_scope('test')
        kv.set('hello', 'saucy')
        kv.set('hello', 'world')
        self.assertEqual(kv.get('hello'), 'world')
        kv.flush()
        kv.unset('hello')
        self.assertEqual(kv.get('hello'), None)
        kv.flush(False)
        self.assertEqual(kv.get('hello'), 'world')


if __name__ == '__main__':
    unittest.main()



#!/usr/bin/python
#
# Copyright 2016 Canonical Ltd
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#  http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os

from mock import MagicMock
from mock import patch

from charmhelpers.contrib.hardware import pci
from tests.contrib.hardware import pci_responses
from tests.contrib.hardware.test_utils import patch_open


def check_device(device, attr_dict):
    equal = (
        device.interface_name == attr_dict["interface_name"] and
        device.mac_address == attr_dict["mac_address"] and
        device.pci_address == attr_dict["pci_address"] and
        device.state == attr_dict["state"]
    )
    return equal


def mocked_subprocess(subproc_map=None):
    def _subproc(cmd, stdin=None):
        for key in pci_responses.COMMANDS.keys():
            if pci_responses.COMMANDS[key] == cmd:
                return subproc_map[key]
            elif pci_responses.COMMANDS[key] == cmd[:-1]:
                return subproc_map[cmd[-1]][key]

    if not subproc_map:
        subproc_map = pci_responses.NET_SETUP
    return _subproc


class mocked_filehandle(object):
    def _setfilename(self, fname, omode):
        self.FILENAME = fname

    def _getfilecontents_read(self):
        try:
            return pci_responses.FILE_CONTENTS[self.FILENAME]
        except KeyError:
            raise OSError()

    def _getfilecontents_readlines(self):
        return pci_responses.FILE_CONTENTS[self.FILENAME].split("\n")


def mocked_globs(path):
    check_path = path.split("*")[0].rstrip("/")
    dirs = []
    for sdir in pci_responses.SYS_TREE:
        if check_path in sdir:
            dirs.append(sdir)
    return dirs


def mocked_islink(link):
    resolved_relpath = mocked_resolve_link(link)
    if pci_responses.SYS_TREE.get(resolved_relpath):
        return True
    else:
        return False


def mocked_resolve_link(link):
    resolved_relpath = None
    for sdir in pci_responses.SYS_TREE:
        if sdir in link:
            rep_dir = "{}/{}".format(
                os.path.dirname(sdir), pci_responses.SYS_TREE[sdir]
            )
            resolved_symlink = link.replace(sdir, rep_dir)
            resolved_relpath = os.path.abspath(resolved_symlink)
    return resolved_relpath


def mocked_realpath(link):
    resolved_link = mocked_resolve_link(link)
    return pci_responses.SYS_TREE[resolved_link]


def mocked_listdir(path):
    return pci_responses.NETDEVS[path]


@patch("pci.cached")
@patch("pci.log")
@patch("pci.subprocess.Popen")
@patch("pci.subprocess.check_output")
@patch("pci.glob.glob")
@patch("pci.os.path.islink")
def pci_devs(
    _osislink, _glob, _check_output, _Popen, _log, _cached, subproc_map=None
):
    _glob.side_effect = mocked_globs
    _osislink.side_effect = mocked_islink
    _check_output.side_effect = mocked_subprocess(subproc_map=subproc_map)

    with patch_open() as (_open, _file), patch(
        "pci.os.path.realpath"
    ) as _realpath:
        super_fh = mocked_filehandle()
        _file.readlines = MagicMock()
        _open.side_effect = super_fh._setfilename
        _file.read.side_effect = super_fh._getfilecontents_read
        _file.readlines.side_effect = super_fh._getfilecontents_readlines
        _realpath.side_effect = mocked_realpath
        devices = pci.PCINetDevices()
    return devices



# Copyright 2014-2015 Canonical Limited.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#  http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import json

from charmhelpers.core.hookenv import relation_id as current_relation_id
from charmhelpers.core.hookenv import (
    is_relation_made,
    relation_ids,
    relation_get as _relation_get,
    local_unit,
    relation_set as _relation_set,
    leader_get as _leader_get,
    leader_set,
    is_leader,
)


"""
This helper provides functions to support use of a peer relation
for basic key/value storage, with the added benefit that all storage
can be replicated across peer units.

Requirement to use:

To use this, the "peer_echo()" method has to be called form the peer
relation's relation-changed hook:

@hooks.hook("cluster-relation-changed") # Adapt the to your peer relation name
def cluster_relation_changed():
    peer_echo()

Once this is done, you can use peer storage from anywhere:

@hooks.hook("some-hook")
def some_hook():
    # You can store and retrieve key/values this way:
    if is_relation_made("cluster"):  # from charmhelpers.core.hookenv
        # There are peers available so we can work with peer storage
        peer_store("mykey", "myvalue")
        value = peer_retrieve("mykey")
        print value
    else:
        print "No peers joind the relation, cannot share key/values :("
"""


def leader_get(attribute=None, rid=None):
    """Wrapper to ensure that settings are migrated from the peer relation.

    This is to support upgrading an environment that does not support
    Juju leadership election to one that does.

    If a setting is not extant in the leader-get but is on the relation-get
    peer rel, it is migrated and marked as such so that it is not re-migrated.
    """
    migration_key = '__leader_get_migrated_settings__'
    if not is_leader():
        return _leader_get(attribute=attribute)

    settings_migrated = False
    leader_settings = _leader_get(attribute=attribute)
    previously_migrated = _leader_get(attribute=migration_key)

    if previously_migrated:
        migrated = set(json.loads(previously_migrated))
    else:
        migrated = set([])

    try:
        if migration_key in leader_settings:
            del leader_settings[migration_key]
    except TypeError:
        pass

    if attribute:
        if attribute in migrated:
            return leader_settings

        # If attribute not present in leader db, check if this unit has set
        # the attribute in the peer relation
        if not leader_settings:
            peer_setting = _relation_get(attribute=attribute, unit=local_unit(),
                                         rid=rid)
            if peer_setting:
                leader_set(settings={attribute: peer_setting})
                leader_settings = peer_setting

        if leader_settings:
            settings_migrated = True
            migrated.add(attribute)
    else:
        r_settings = _relation_get(unit=local_unit(), rid=rid)
        if r_settings:
            for key in set(r_settings.keys()).difference(migrated):
                # Leader setting wins
                if not leader_settings.get(key):
                    leader_settings[key] = r_settings[key]

                settings_migrated = True
                migrated.add(key)

            if settings_migrated:
                leader_set(**leader_settings)

    if migrated and settings_migrated:
        migrated = json.dumps(list(migrated))
        leader_set(settings={migration_key: migrated})

    return leader_settings


def relation_set(relation_id=None, relation_settings=None, **kwargs):
    """Attempt to use leader-set if supported in the current version of Juju,
    otherwise falls back on relation-set.

    Note that we only attempt to use leader-set if the provided relation_id is
    a peer relation id or no relation id is provided (in which case we assume
    we are within the peer relation context).
    """
    try:
        if relation_id in relation_ids('cluster'):
            return leader_set(settings=relation_settings, **kwargs)
        else:
            raise NotImplementedError
    except NotImplementedError:
        return _relation_set(relation_id=relation_id,
                             relation_settings=relation_settings, **kwargs)


def relation_get(attribute=None, unit=None, rid=None):
    """Attempt to use leader-get if supported in the current version of Juju,
    otherwise falls back on relation-get.

    Note that we only attempt to use leader-get if the provided rid is a peer
    relation id or no relation id is provided (in which case we assume we are
    within the peer relation context).
    """
    try:
        if rid in relation_ids('cluster'):
            return leader_get(attribute, rid)
        else:
            raise NotImplementedError
    except NotImplementedError:
        return _relation_get(attribute=attribute, rid=rid, unit=unit)


def peer_retrieve(key, relation_name='cluster'):
    """Retrieve a named key from peer relation `relation_name`."""
    cluster_rels = relation_ids(relation_name)
    if len(cluster_rels) > 0:
        cluster_rid = cluster_rels[0]
        return relation_get(attribute=key, rid=cluster_rid,
                            unit=local_unit())
    else:
        raise ValueError('Unable to detect'
                         'peer relation {}'.format(relation_name))


def peer_retrieve_by_prefix(prefix, relation_name='cluster', delimiter='_',
                            inc_list=None, exc_list=None):
    """ Retrieve k/v pairs given a prefix and filter using {inc,exc}_list """
    inc_list = inc_list if inc_list else []
    exc_list = exc_list if exc_list else []
    peerdb_settings = peer_retrieve('-', relation_name=relation_name)
    matched = {}
    if peerdb_settings is None:
        return matched
    for k, v in peerdb_settings.items():
        full_prefix = prefix + delimiter
        if k.startswith(full_prefix):
            new_key = k.replace(full_prefix, '')
            if new_key in exc_list:
                continue
            if new_key in inc_list or len(inc_list) == 0:
                matched[new_key] = v
    return matched


def peer_store(key, value, relation_name='cluster'):
    """Store the key/value pair on the named peer relation `relation_name`."""
    cluster_rels = relation_ids(relation_name)
    if len(cluster_rels) > 0:
        cluster_rid = cluster_rels[0]
        relation_set(relation_id=cluster_rid,
                     relation_settings={key: value})
    else:
        raise ValueError('Unable to detect '
                         'peer relation {}'.format(relation_name))


def peer_echo(includes=None, force=False):
    """Echo filtered attributes back onto the same relation for storage.

    This is a requirement to use the peerstorage module - it needs to be called
    from the peer relation's changed hook.

    If Juju leader support exists this will be a noop unless force is True.
    """
    try:
        is_leader()
    except NotImplementedError:
        pass
    else:
        if not force:
            return  # NOOP if leader-election is supported

    # Use original non-leader calls
    relation_get = _relation_get
    relation_set = _relation_set

    rdata = relation_get()
    echo_data = {}
    if includes is None:
        echo_data = rdata.copy()
        for ex in ['private-address', 'public-address']:
            if ex in echo_data:
                echo_data.pop(ex)
    else:
        for attribute, value in rdata.items():
            for include in includes:
                if include in attribute:
                    echo_data[attribute] = value
    if len(echo_data) > 0:
        relation_set(relation_settings=echo_data)


def peer_store_and_set(relation_id=None, peer_relation_name='cluster',
                       peer_store_fatal=False, relation_settings=None,
                       delimiter='_', **kwargs):
    """Store passed-in arguments both in argument relation and in peer storage.

    It functions like doing relation_set() and peer_store() at the same time,
    with the same data.

    @param relation_id: the id of the relation to store the data on. Defaults
                        to the current relation.
    @param peer_store_fatal: Set to True, the function will raise an exception
                             should the peer storage not be available."""

    relation_settings = relation_settings if relation_settings else {}
    relation_set(relation_id=relation_id,
                 relation_settings=relation_settings,
                 **kwargs)
    if is_relation_made(peer_relation_name):
        items = dict(list(kwargs.items()) + list(relation_settings.items()))
        for key, value in items.items():
            key_prefix = relation_id or current_relation_id()
            peer_store(key_prefix + delimiter + key,
                       value,
                       relation_name=peer_relation_name)
    else:
        if peer_store_fatal:
            raise ValueError('Unable to detect '
                             'peer relation {}'.format(peer_relation_name))



import subprocess


def persistent_modprobe(module):
    """Load a kernel module and configure for auto-load on reboot."""
    with open('/etc/modules', 'r+') as modules:
        if module not in modules.read():
            modules.write(module + "\n")


def update_initramfs(version='all'):
    """Updates an initramfs image."""
    return subprocess.check_call(["update-initramfs", "-k", version, "-u"])



# Copyright 2016 Canonical Limited.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#  http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from unittest import TestCase

from mock import call
from mock import MagicMock
from mock import patch

from charmhelpers.contrib.hardening.audits import apt
from charmhelpers.fetch import ubuntu_apt_pkg as apt_pkg
from charmhelpers.core import hookenv


class RestrictedPackagesTestCase(TestCase):
    def setUp(self):
        super(RestrictedPackagesTestCase, self).setUp()

    def create_package(self, name, virtual=False):
        pkg = MagicMock()
        pkg.name = name
        pkg.current_ver = '2.0'
        if virtual:
            pkgver = MagicMock()
            pkgver.parent_pkg = self.create_package('foo')
            pkg.provides_list = [('virtualfoo', None, pkgver)]
            resp = {
                'has_provides': True,
                'has_versions': False,
            }
            pkg.get.side_effect = resp.get

        return pkg

    @patch.object(apt, 'apt_cache')
    @patch.object(apt, 'apt_purge')
    @patch.object(apt, 'log', lambda *args, **kwargs: None)
    def test_ensure_compliance(self, mock_purge, mock_apt_cache):
        pkg = self.create_package('bar')
        mock_apt_cache.return_value = {'bar': pkg}

        audit = apt.RestrictedPackages(pkgs=['bar'])
        audit.ensure_compliance()
        mock_purge.assert_has_calls([call(pkg.name)])

    @patch.object(apt, 'apt_purge')
    @patch.object(apt, 'apt_cache')
    @patch.object(apt, 'log', lambda *args, **kwargs: None)
    def test_apt_harden_virtual_package(self, mock_apt_cache, mock_apt_purge):
        vpkg = self.create_package('virtualfoo', virtual=True)
        mock_apt_cache.return_value = {'foo': vpkg}
        audit = apt.RestrictedPackages(pkgs=['foo'])
        audit.ensure_compliance()
        self.assertTrue(mock_apt_cache.called)
        mock_apt_purge.assert_has_calls([call('foo')])


class AptConfigTestCase(TestCase):

    @patch.object(apt, 'apt_pkg')
    def test_ensure_compliance(self, mock_apt_pkg):
        mock_apt_pkg.init.return_value = None
        mock_apt_pkg.config.side_effect = {}
        mock_apt_pkg.config.get.return_value = None
        audit = apt.AptConfig([{'key': 'APT::Get::AllowUnauthenticated',
                                'expected': 'false'}])
        audit.ensure_compliance()
        self.assertTrue(mock_apt_pkg.config.get.called)

    @patch.object(hookenv, 'log')
    def test_verify_config(self, mock_log):
        cfg = apt_pkg.config
        key, value = list(cfg.items())[0]
        audit = apt.AptConfig([{"key": key, "expected": value}])
        audit.ensure_compliance()
        self.assertFalse(mock_log.called)



"""Tests for the commandant code that analyzes a function signature to
determine the parameters to argparse."""

from unittest import TestCase
from mock import (
    patch,
    MagicMock,
    ANY,
)
import json
from pprint import pformat
import yaml
import csv

from io import StringIO

from charmhelpers import cli


class SubCommandTest(TestCase):
    """Test creation of subcommands"""

    def setUp(self):
        super(SubCommandTest, self).setUp()
        self.cl = cli.CommandLine()

    @patch('sys.exit')
    def test_subcommand_wrapper(self, _sys_exit):
        """Test function name detection"""
        @self.cl.subcommand()
        def payload():
            "A function that does work."
            pass
        args = self.cl.argument_parser.parse_args(['payload'])
        self.assertEqual(args.func, payload)
        self.assertEqual(_sys_exit.mock_calls, [])

    @patch('sys.exit')
    def test_subcommand_wrapper_bogus_arguments(self, _sys_exit):
        """Test function name detection"""
        @self.cl.subcommand()
        def payload():
            "A function that does work."
            pass
        with self.assertRaises(TypeError):
            with patch("sys.argv", "tests deliberately bad input".split()):
                with patch("sys.stderr"):
                    self.cl.argument_parser.parse_args()
        _sys_exit.assert_called_once_with(2)

    @patch('sys.exit')
    def test_subcommand_wrapper_cmdline_options(self, _sys_exit):
        """Test detection of positional arguments and optional parameters."""
        @self.cl.subcommand()
        def payload(x, y=None):
            "A function that does work."
            return x
        args = self.cl.argument_parser.parse_args(['payload', 'positional', '--y=optional'])
        self.assertEqual(args.func, payload)
        self.assertEqual(args.x, 'positional')
        self.assertEqual(args.y, 'optional')
        self.assertEqual(_sys_exit.mock_calls, [])

    @patch('sys.exit')
    def test_subcommand_builder(self, _sys_exit):
        def noop(z):
            pass

        @self.cl.subcommand_builder('payload', description="A subcommand")
        def payload_command(subparser):
            subparser.add_argument('-z', action='store_true')
            return noop

        args = self.cl.argument_parser.parse_args(['payload', '-z'])
        self.assertEqual(args.func, noop)
        self.assertTrue(args.z)
        self.assertFalse(_sys_exit.called)

    def test_subcommand_builder_bogus_wrapped_args(self):
        with self.assertRaises(TypeError):
            @self.cl.subcommand_builder('payload', description="A subcommand")
            def payload_command(subparser, otherarg):
                pass

    def test_run(self):
        self.bar_called = False

        @self.cl.subcommand()
        def bar(x, y=None, *vargs):
            "A function that does work."
            self.assertEqual(x, 'baz')
            self.assertEqual(y, 'why')
            self.assertEqual(vargs, ('mux', 'zob'))
            self.bar_called = True
            return "qux"

        args = ['chlp', 'bar', '--y', 'why', 'baz', 'mux', 'zob']
        self.cl.formatter = MagicMock()
        with patch("sys.argv", args):
            with patch("charmhelpers.core.unitdata._KV") as _KV:
                self.cl.run()
                assert _KV.flush.called
        self.assertTrue(self.bar_called)
        self.cl.formatter.format_output.assert_called_once_with('qux', ANY)

    def test_no_output(self):
        self.bar_called = False

        @self.cl.subcommand()
        @self.cl.no_output
        def bar(x, y=None, *vargs):
            "A function that does work."
            self.bar_called = True
            return "qux"

        args = ['foo', 'bar', 'baz']
        self.cl.formatter = MagicMock()
        with patch("sys.argv", args):
            self.cl.run()
        self.assertTrue(self.bar_called)
        self.cl.formatter.format_output.assert_called_once_with('', ANY)

    def test_test_command(self):
        self.bar_called = False
        self.bar_result = True

        @self.cl.subcommand()
        @self.cl.test_command
        def bar(x, y=None, *vargs):
            "A function that does work."
            self.bar_called = True
            return self.bar_result

        args = ['foo', 'bar', 'baz']
        self.cl.formatter = MagicMock()
        with patch("sys.argv", args):
            self.cl.run()
        self.assertTrue(self.bar_called)
        self.assertEqual(self.cl.exit_code, 0)
        self.cl.formatter.format_output.assert_called_once_with('', ANY)

        self.bar_result = False
        with patch("sys.argv", args):
            self.cl.run()
        self.assertEqual(self.cl.exit_code, 1)


class OutputFormatterTest(TestCase):
    def setUp(self):
        super(OutputFormatterTest, self).setUp()
        self.expected_formats = (
            "raw",
            "json",
            "py",
            "yaml",
            "csv",
            "tab",
        )
        self.outfile = StringIO()
        self.of = cli.OutputFormatter(outfile=self.outfile)
        self.output_data = {"this": "is", "some": 1, "data": dict()}

    def test_supports_formats(self):
        self.assertEqual(sorted(self.expected_formats),
                         sorted(self.of.supported_formats))

    def test_adds_arguments(self):
        ap = MagicMock()
        arg_group = MagicMock()
        add_arg = MagicMock()
        arg_group.add_argument = add_arg
        ap.add_mutually_exclusive_group.return_value = arg_group
        self.of.add_arguments(ap)

        self.assertTrue(add_arg.called)

        for call_args in add_arg.call_args_list:
            if "--format" in call_args[0]:
                self.assertEqual(sorted(call_args[1]['choices']),
                                 sorted(self.expected_formats))
                self.assertEqual(call_args[1]['default'], 'raw')
                break
        else:
            print(arg_group.call_args_list)
            self.fail("No --format argument was created")

        all_args = [c[0][0] for c in add_arg.call_args_list]
        all_args.extend([c[0][1] for c in add_arg.call_args_list if len(c[0]) > 1])
        for fmt in self.expected_formats:
            self.assertIn("-{}".format(fmt[0]), all_args)
            self.assertIn("--{}".format(fmt), all_args)

    def test_outputs_raw(self):
        self.of.raw(self.output_data)
        self.outfile.seek(0)
        self.assertEqual(self.outfile.read(), str(self.output_data))

    def test_outputs_json(self):
        self.of.json(self.output_data)
        self.outfile.seek(0)
        self.assertEqual(self.outfile.read(), json.dumps(self.output_data))

    def test_outputs_py(self):
        self.of.py(self.output_data)
        self.outfile.seek(0)
        self.assertEqual(self.outfile.read(), pformat(self.output_data) + "\n")

    def test_outputs_yaml(self):
        self.of.yaml(self.output_data)
        self.outfile.seek(0)
        self.assertEqual(self.outfile.read(), yaml.dump(self.output_data))

    def test_outputs_csv(self):
        sample = StringIO()
        writer = csv.writer(sample)
        writer.writerows(self.output_data)
        sample.seek(0)
        self.of.csv(self.output_data)
        self.outfile.seek(0)
        self.assertEqual(self.outfile.read(), sample.read())

    def test_outputs_tab(self):
        sample = StringIO()
        writer = csv.writer(sample, dialect=csv.excel_tab)
        writer.writerows(self.output_data)
        sample.seek(0)
        self.of.tab(self.output_data)
        self.outfile.seek(0)
        self.assertEqual(self.outfile.read(), sample.read())

    def test_formats_output(self):
        for format in self.expected_formats:
            mock_f = MagicMock()
            setattr(self.of, format, mock_f)
            self.of.format_output(self.output_data, format)
            mock_f.assert_called_with(self.output_data)



# Copyright 2019-2021 Canonical Ltd
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#  http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import collections
import contextlib
import os
import shutil
import yaml
import zipfile

import charmhelpers
import charmhelpers.core.hookenv as hookenv
import charmhelpers.core.host as ch_host

# Features provided by this module:

"""
Policy.d helper functions
=========================

The functions in this module are designed, as a set, to provide an easy-to-use
set of hooks for classic charms to add in /etc/<service-name>/policy.d/
directory override YAML files.

(For charms.openstack charms, a mixin class is provided for this
functionality).

In order to "hook" this functionality into a (classic) charm, two functions are
provided:

    maybe_do_policyd_overrides(openstack_release,
                               service,
                               blacklist_paths=none,
                               blacklist_keys=none,
                               template_function=none,
                               restart_handler=none)

    maybe_do_policyd_overrides_on_config_changed(openstack_release,
                                                 service,
                                                 blacklist_paths=None,
                                                 blacklist_keys=None,
                                                 template_function=None,
                                                 restart_handler=None

(See the docstrings for details on the parameters)

The functions should be called from the install and upgrade hooks in the charm.
The `maybe_do_policyd_overrides_on_config_changed` function is designed to be
called on the config-changed hook, in that it does an additional check to
ensure that an already overridden policy.d in an upgrade or install hooks isn't
repeated.

In order the *enable* this functionality, the charm's install, config_changed,
and upgrade_charm hooks need to be modified, and a new config option (see
below) needs to be added.  The README for the charm should also be updated.

Examples from the keystone charm are:

@hooks.hook('install.real')
@harden()
def install():
    ...
    # call the policy overrides handler which will install any policy overrides
    maybe_do_policyd_overrides(os_release('keystone'), 'keystone')


@hooks.hook('config-changed')
@restart_on_change(restart_map(), restart_functions=restart_function_map())
@harden()
def config_changed():
    ...
    # call the policy overrides handler which will install any policy overrides
    maybe_do_policyd_overrides_on_config_changed(os_release('keystone'),
                                                 'keystone')

@hooks.hook('upgrade-charm')
@restart_on_change(restart_map(), stopstart=True)
@harden()
def upgrade_charm():
    ...
    # call the policy overrides handler which will install any policy overrides
    maybe_do_policyd_overrides(os_release('keystone'), 'keystone')

Status Line
===========

The workload status code in charm-helpers has been modified to detect if
policy.d override code has been incorporated into the charm by checking for the
new config variable (in the config.yaml).  If it has been, then the workload
status line will automatically show "PO:" at the beginning of the workload
status for that unit/service if the config option is set.  If the policy
override is broken, the "PO (broken):" will be shown.  No changes to the charm
(apart from those already mentioned) are needed to enable this functionality.
(charms.openstack charms also get this functionality, but please see that
library for further details).
"""

# The config.yaml for the charm should contain the following for the config
# option:

"""
  use-policyd-override:
    type: boolean
    default: False
    description: |
      If True then use the resource file named 'policyd-override' to install
      override YAML files in the service's policy.d directory.  The resource
      file should be a ZIP file containing at least one yaml file with a .yaml
      or .yml extension.  If False then remove the overrides.
"""

# The metadata.yaml for the charm should contain the following:
"""
resources:
  policyd-override:
    type: file
    filename: policyd-override.zip
    description: The policy.d overrides file
"""

# The README for the charm should contain the following:
"""
Policy Overrides
----------------

This feature allows for policy overrides using the `policy.d` directory.  This
is an **advanced** feature and the policies that the OpenStack service supports
should be clearly and unambiguously understood before trying to override, or
add to, the default policies that the service uses.  The charm also has some
policy defaults.  They should also be understood before being overridden.

> **Caution**: It is possible to break the system (for tenants and other
  services) if policies are incorrectly applied to the service.

Policy overrides are YAML files that contain rules that will add to, or
override, existing policy rules in the service.  The `policy.d` directory is
a place to put the YAML override files.  This charm owns the
`/etc/keystone/policy.d` directory, and as such, any manual changes to it will
be overwritten on charm upgrades.

Overrides are provided to the charm using a Juju resource called
`policyd-override`.  The resource is a ZIP file.  This file, say
`overrides.zip`, is attached to the charm by:


    juju attach-resource <charm-name> policyd-override=overrides.zip

The policy override is enabled in the charm using:

    juju config <charm-name> use-policyd-override=true

When `use-policyd-override` is `True` the status line of the charm will be
prefixed with `PO:` indicating that policies have been overridden.  If the
installation of the policy override YAML files failed for any reason then the
status line will be prefixed with `PO (broken):`.  The log file for the charm
will indicate the reason.  No policy override files are installed if the `PO
(broken):` is shown.  The status line indicates that the overrides are broken,
not that the policy for the service has failed. The policy will be the defaults
for the charm and service.

Policy overrides on one service may affect the functionality of another
service. Therefore, it may be necessary to provide policy overrides for
multiple service charms to achieve a consistent set of policies across the
OpenStack system.  The charms for the other services that may need overrides
should be checked to ensure that they support overrides before proceeding.
"""

POLICYD_VALID_EXTS = ['.yaml', '.yml', '.j2', '.tmpl', '.tpl']
POLICYD_TEMPLATE_EXTS = ['.j2', '.tmpl', '.tpl']
POLICYD_RESOURCE_NAME = "policyd-override"
POLICYD_CONFIG_NAME = "use-policyd-override"
POLICYD_SUCCESS_FILENAME = "policyd-override-success"
POLICYD_LOG_LEVEL_DEFAULT = hookenv.INFO
POLICYD_ALWAYS_BLACKLISTED_KEYS = ("admin_required", "cloud_admin")


class BadPolicyZipFile(Exception):

    def __init__(self, log_message):
        self.log_message = log_message

    def __str__(self):
        return self.log_message


class BadPolicyYamlFile(Exception):

    def __init__(self, log_message):
        self.log_message = log_message

    def __str__(self):
        return self.log_message


def is_policyd_override_valid_on_this_release(openstack_release):
    """Check that the charm is running on at least Ubuntu Xenial, and at
    least the queens release.

    :param openstack_release: the release codename that is installed.
    :type openstack_release: str
    :returns: True if okay
    :rtype: bool
    """
    # NOTE(ajkavanagh) circular import!  This is because the status message
    # generation code in utils has to call into this module, but this function
    # needs the CompareOpenStackReleases() function.  The only way to solve
    # this is either to put ALL of this module into utils, or refactor one or
    # other of the CompareOpenStackReleases or status message generation code
    # into a 3rd module.
    import charmhelpers.contrib.openstack.utils as ch_utils
    return ch_utils.CompareOpenStackReleases(openstack_release) >= 'queens'


def maybe_do_policyd_overrides(openstack_release,
                               service,
                               blacklist_paths=None,
                               blacklist_keys=None,
                               template_function=None,
                               restart_handler=None,
                               user=None,
                               group=None,
                               config_changed=False):
    """If the config option is set, get the resource file and process it to
    enable the policy.d overrides for the service passed.

    The param `openstack_release` is required as the policyd overrides feature
    is only supported on openstack_release "queens" or later, and on ubuntu
    "xenial" or later.  Prior to these versions, this feature is a NOP.

    The optional template_function is a function that accepts a string and has
    an opportunity to modify the loaded file prior to it being read by
    yaml.safe_load().  This allows the charm to perform "templating" using
    charm derived data.

    The param blacklist_paths are paths (that are in the service's policy.d
    directory that should not be touched).

    The param blacklist_keys are keys that must not appear in the yaml file.
    If they do, then the whole policy.d file fails.

    The yaml file extracted from the resource_file (which is a zipped file) has
    its file path reconstructed.  This, also, must not match any path in the
    black list.

    The param restart_handler is an optional Callable that is called to perform
    the service restart if the policy.d file is changed.  This should normally
    be None as oslo.policy automatically picks up changes in the policy.d
    directory.  However, for any services where this is buggy then a
    restart_handler can be used to force the policy.d files to be read.

    If the config_changed param is True, then the handling is slightly
    different: It will only perform the policyd overrides if the config is True
    and the success file doesn't exist.  Otherwise, it does nothing as the
    resource file has already been processed.

    :param openstack_release: The openstack release that is installed.
    :type openstack_release: str
    :param service: the service name to construct the policy.d directory for.
    :type service: str
    :param blacklist_paths: optional list of paths to leave alone
    :type blacklist_paths: Union[None, List[str]]
    :param blacklist_keys: optional list of keys that mustn't appear in the
                           yaml file's
    :type blacklist_keys: Union[None, List[str]]
    :param template_function: Optional function that can modify the string
                              prior to being processed as a Yaml document.
    :type template_function: Union[None, Callable[[str], str]]
    :param restart_handler: The function to call if the service should be
                            restarted.
    :type restart_handler: Union[None, Callable[]]
    :param user: The user to create/write files/directories as
    :type user: Union[None, str]
    :param group: the group to create/write files/directories as
    :type group: Union[None, str]
    :param config_changed: Set to True for config_changed hook.
    :type config_changed: bool
    """
    _user = service if user is None else user
    _group = service if group is None else group
    if not is_policyd_override_valid_on_this_release(openstack_release):
        return
    hookenv.log("Running maybe_do_policyd_overrides",
                level=POLICYD_LOG_LEVEL_DEFAULT)
    config = hookenv.config()
    try:
        if not config.get(POLICYD_CONFIG_NAME, False):
            clean_policyd_dir_for(service,
                                  blacklist_paths,
                                  user=_user,
                                  group=_group)
            if (os.path.isfile(_policy_success_file()) and
                    restart_handler is not None and
                    callable(restart_handler)):
                restart_handler()
            remove_policy_success_file()
            return
    except Exception as e:
        hookenv.log("... ERROR: Exception is: {}".format(str(e)),
                    level=POLICYD_CONFIG_NAME)
        import traceback
        hookenv.log(traceback.format_exc(), level=POLICYD_LOG_LEVEL_DEFAULT)
        return
    # if the policyd overrides have been performed when doing config_changed
    # just return
    if config_changed and is_policy_success_file_set():
        hookenv.log("... already setup, so skipping.",
                    level=POLICYD_LOG_LEVEL_DEFAULT)
        return
    # from now on it should succeed; if it doesn't then status line will show
    # broken.
    resource_filename = get_policy_resource_filename()
    restart = process_policy_resource_file(
        resource_filename, service, blacklist_paths, blacklist_keys,
        template_function)
    if restart and restart_handler is not None and callable(restart_handler):
        restart_handler()


@charmhelpers.deprecate("Use maybe_do_policyd_overrides instead")
def maybe_do_policyd_overrides_on_config_changed(*args, **kwargs):
    """This function is designed to be called from the config changed hook.

    DEPRECATED: please use maybe_do_policyd_overrides() with the param
    `config_changed` as `True`.

    See maybe_do_policyd_overrides() for more details on the params.
    """
    if 'config_changed' not in kwargs.keys():
        kwargs['config_changed'] = True
    return maybe_do_policyd_overrides(*args, **kwargs)


def get_policy_resource_filename():
    """Function to extract the policy resource filename

    :returns: The filename of the resource, if set, otherwise, if an error
               occurs, then None is returned.
    :rtype: Union[str, None]
    """
    try:
        return hookenv.resource_get(POLICYD_RESOURCE_NAME)
    except Exception:
        return None


@contextlib.contextmanager
def open_and_filter_yaml_files(filepath, has_subdirs=False):
    """Validate that the filepath provided is a zip file and contains at least
    one (.yaml|.yml) file, and that the files are not duplicated when the zip
    file is flattened.  Note that the yaml files are not checked.  This is the
    first stage in validating the policy zipfile; individual yaml files are not
    checked for validity or black listed keys.

    If the has_subdirs param is True, then the files are flattened to the first
    directory, and the files in the root are ignored.

    An example of use is:

        with open_and_filter_yaml_files(some_path) as zfp, g:
            for zipinfo in g:
                # do something with zipinfo ...

    :param filepath: a filepath object that can be opened by zipfile
    :type filepath: Union[AnyStr, os.PathLike[AntStr]]
    :param has_subdirs: Keep first level of subdirectories in yaml file.
    :type has_subdirs: bool
    :returns: (zfp handle,
               a generator of the (name, filename, ZipInfo object) tuples) as a
               tuple.
    :rtype: ContextManager[(zipfile.ZipFile,
                            Generator[(name, str, str, zipfile.ZipInfo)])]
    :raises: zipfile.BadZipFile
    :raises: BadPolicyZipFile if duplicated yaml or missing
    :raises: IOError if the filepath is not found
    """
    with zipfile.ZipFile(filepath, 'r') as zfp:
        # first pass through; check for duplicates and at least one yaml file.
        names = collections.defaultdict(int)
        yamlfiles = _yamlfiles(zfp, has_subdirs)
        for name, _, _, _ in yamlfiles:
            names[name] += 1
        # There must be at least 1 yaml file.
        if len(names.keys()) == 0:
            raise BadPolicyZipFile("contains no yaml files with {} extensions."
                                   .format(", ".join(POLICYD_VALID_EXTS)))
        # There must be no duplicates
        duplicates = [n for n, c in names.items() if c > 1]
        if duplicates:
            raise BadPolicyZipFile("{} have duplicates in the zip file."
                                   .format(", ".join(duplicates)))
        # Finally, let's yield the generator
        yield (zfp, yamlfiles)


def _yamlfiles(zipfile, has_subdirs=False):
    """Helper to get a yaml file (according to POLICYD_VALID_EXTS extensions)
    and the infolist item from a zipfile.

    If the `has_subdirs` param is True, the the only yaml files that have a
    directory component are read, and then first part of the directory
    component is kept, along with the filename in the name.  e.g. an entry with
    a filename of:

        compute/someotherdir/override.yaml

    is returned as:

        compute/override, yaml, override.yaml, <ZipInfo object>

    This is to help with the special, additional, processing that the dashboard
    charm requires.

    :param zipfile: the zipfile to read zipinfo items from
    :type zipfile: zipfile.ZipFile
    :param has_subdirs: Keep first level of subdirectories in yaml file.
    :type has_subdirs: bool
    :returns: generator of (name, ext, filename, info item) for each
              self-identified yaml file.
    :rtype: List[(str, str, str, zipfile.ZipInfo)]
    """
    files = []
    for infolist_item in zipfile.infolist():
        try:
            if infolist_item.is_dir():
                continue
        except AttributeError:
            # fallback to "old" way to determine dir entry for pre-py36
            if infolist_item.filename.endswith('/'):
                continue
        _dir, name_ext = os.path.split(infolist_item.filename)
        name, ext = os.path.splitext(name_ext)
        if has_subdirs and _dir != "":
            name = os.path.join(_dir.split(os.path.sep)[0], name)
        ext = ext.lower()
        if ext and ext in POLICYD_VALID_EXTS:
            files.append((name, ext, name_ext, infolist_item))
    return files


def read_and_validate_yaml(stream_or_doc, blacklist_keys=None):
    """Read, validate and return the (first) yaml document from the stream.

    The doc is read, and checked for a yaml file.  The the top-level keys are
    checked against the blacklist_keys provided.  If there are problems then an
    Exception is raised.  Otherwise the yaml document is returned as a Python
    object that can be dumped back as a yaml file on the system.

    The yaml file must only consist of a str:str mapping, and if not then the
    yaml file is rejected.

    :param stream_or_doc: the file object to read the yaml from
    :type stream_or_doc: Union[AnyStr, IO[AnyStr]]
    :param blacklist_keys: Any keys, which if in the yaml file, should cause
        and error.
    :type blacklisted_keys: Union[None, List[str]]
    :returns: the yaml file as a python document
    :rtype: Dict[str, str]
    :raises: yaml.YAMLError if there is a problem with the document
    :raises: BadPolicyYamlFile if file doesn't look right or there are
             blacklisted keys in the file.
    """
    blacklist_keys = blacklist_keys or []
    blacklist_keys.append(POLICYD_ALWAYS_BLACKLISTED_KEYS)
    doc = yaml.safe_load(stream_or_doc)
    if not isinstance(doc, dict):
        raise BadPolicyYamlFile("doesn't look like a policy file?")
    keys = set(doc.keys())
    blacklisted_keys_present = keys.intersection(blacklist_keys)
    if blacklisted_keys_present:
        raise BadPolicyYamlFile("blacklisted keys {} present."
                                .format(", ".join(blacklisted_keys_present)))
    if not all(isinstance(k, str) for k in keys):
        raise BadPolicyYamlFile("keys in yaml aren't all strings?")
    # check that the dictionary looks like a mapping of str to str
    if not all(isinstance(v, str) for v in doc.values()):
        raise BadPolicyYamlFile("values in yaml aren't all strings?")
    return doc


def policyd_dir_for(service):
    """Return the policy directory for the named service.

    :param service: str
    :returns: the policy.d override directory.
    :rtype: os.PathLike[str]
    """
    return os.path.join("/", "etc", service, "policy.d")


def clean_policyd_dir_for(service, keep_paths=None, user=None, group=None):
    """Clean out the policyd directory except for items that should be kept.

    The keep_paths, if used, should be set to the full path of the files that
    should be kept in the policyd directory for the service.  Note that the
    service name is passed in, and then the policyd_dir_for() function is used.
    This is so that a coding error doesn't result in a sudden deletion of the
    charm (say).

    :param service: the service name to use to construct the policy.d dir.
    :type service: str
    :param keep_paths: optional list of paths to not delete.
    :type keep_paths: Union[None, List[str]]
    :param user: The user to create/write files/directories as
    :type user: Union[None, str]
    :param group: the group to create/write files/directories as
    :type group: Union[None, str]
    """
    _user = service if user is None else user
    _group = service if group is None else group
    keep_paths = keep_paths or []
    path = policyd_dir_for(service)
    hookenv.log("Cleaning path: {}".format(path), level=hookenv.DEBUG)
    if not os.path.exists(path):
        ch_host.mkdir(path, owner=_user, group=_group, perms=0o775)
    for direntry in os.scandir(path):
        # see if the path should be kept.
        if direntry.path in keep_paths:
            continue
        # we remove any directories; it's ours and there shouldn't be any
        if direntry.is_dir():
            shutil.rmtree(direntry.path)
        else:
            os.remove(direntry.path)


def maybe_create_directory_for(path, user, group):
    """For the filename 'path', ensure that the directory for that path exists.

    Note that if the directory already exists then the permissions are NOT
    changed.

    :param path: the filename including the path to it.
    :type path: str
    :param user: the user to create the directory as
    :param group: the group to create the directory as
    """
    _dir, _ = os.path.split(path)
    if not os.path.exists(_dir):
        ch_host.mkdir(_dir, owner=user, group=group, perms=0o775)


def path_for_policy_file(service, name):
    """Return the full path for a policy.d file that will be written to the
    service's policy.d directory.

    It is constructed using policyd_dir_for(), the name and the ".yaml"
    extension.

    For horizon, for example, it's a bit more complicated.  The name param is
    actually "override_service_dir/a_name", where target_service needs to be
    one the allowed horizon override services.  This translation and check is
    done in the _yamlfiles() function.

    :param service: the service name
    :type service: str
    :param name: the name for the policy override
    :type name: str
    :returns: the full path name for the file
    :rtype: os.PathLike[str]
    """
    return os.path.join(policyd_dir_for(service), name + ".yaml")


def _policy_success_file():
    """Return the file name for a successful drop of policy.d overrides

    :returns: the path name for the file.
    :rtype: str
    """
    return os.path.join(hookenv.charm_dir(), POLICYD_SUCCESS_FILENAME)


def remove_policy_success_file():
    """Remove the file that indicates successful policyd override."""
    try:
        os.remove(_policy_success_file())
    except Exception:
        pass


def set_policy_success_file():
    """Set the file that indicates successful policyd override."""
    open(_policy_success_file(), "w").close()


def is_policy_success_file_set():
    """Returns True if the policy success file has been set.

    This indicates that policies are overridden and working properly.

    :returns: True if the policy file is set
    :rtype: bool
    """
    return os.path.isfile(_policy_success_file())


def policyd_status_message_prefix():
    """Return the prefix str for the status line.

    "PO:" indicating that the policy overrides are in place, or "PO (broken):"
    if the policy is supposed to be working but there is no success file.

    :returns: the prefix
    :rtype: str
    """
    if is_policy_success_file_set():
        return "PO:"
    return "PO (broken):"


def process_policy_resource_file(resource_file,
                                 service,
                                 blacklist_paths=None,
                                 blacklist_keys=None,
                                 template_function=None,
                                 preserve_topdir=False,
                                 preprocess_filename=None,
                                 user=None,
                                 group=None):
    """Process the resource file (which should contain at least one yaml file)
    and write those files to the service's policy.d directory.

    The optional template_function is a function that accepts a python
    string and has an opportunity to modify the document
    prior to it being read by the yaml.safe_load() function and written to
    disk. Note that this function does *not* say how the templating is done -
    this is up to the charm to implement its chosen method.

    The param blacklist_paths are paths (that are in the service's policy.d
    directory that should not be touched).

    The param blacklist_keys are keys that must not appear in the yaml file.
    If they do, then the whole policy.d file fails.

    The yaml file extracted from the resource_file (which is a zipped file) has
    its file path reconstructed.  This, also, must not match any path in the
    black list.

    The yaml filename can be modified in two ways.  If the `preserve_topdir`
    param is True, then files will be flattened to the top dir.  This allows
    for creating sets of files that can be grouped into a single level tree
    structure.

    Secondly, if the `preprocess_filename` param is not None and callable()
    then the name is passed to that function for preprocessing before being
    converted to the end location.  This is to allow munging of the filename
    prior to being tested for a blacklist path.

    If any error occurs, then the policy.d directory is cleared, the error is
    written to the log, and the status line will eventually show as failed.

    :param resource_file: The zipped file to open and extract yaml files form.
    :type resource_file: Union[AnyStr, os.PathLike[AnyStr]]
    :param service: the service name to construct the policy.d directory for.
    :type service: str
    :param blacklist_paths: optional list of paths to leave alone
    :type blacklist_paths: Union[None, List[str]]
    :param blacklist_keys: optional list of keys that mustn't appear in the
                           yaml file's
    :type blacklist_keys: Union[None, List[str]]
    :param template_function: Optional function that can modify the yaml
                              document.
    :type template_function: Union[None, Callable[[AnyStr], AnyStr]]
    :param preserve_topdir: Keep the toplevel subdir
    :type preserve_topdir: bool
    :param preprocess_filename: Optional function to use to process filenames
                                extracted from the resource file.
    :type preprocess_filename: Union[None, Callable[[AnyStr]. AnyStr]]
    :param user: The user to create/write files/directories as
    :type user: Union[None, str]
    :param group: the group to create/write files/directories as
    :type group: Union[None, str]
    :returns: True if the processing was successful, False if not.
    :rtype: boolean
    """
    hookenv.log("Running process_policy_resource_file", level=hookenv.DEBUG)
    blacklist_paths = blacklist_paths or []
    completed = False
    _preprocess = None
    if preprocess_filename is not None and callable(preprocess_filename):
        _preprocess = preprocess_filename
    _user = service if user is None else user
    _group = service if group is None else group
    try:
        with open_and_filter_yaml_files(
                resource_file, preserve_topdir) as (zfp, gen):
            # first clear out the policy.d directory and clear success
            remove_policy_success_file()
            clean_policyd_dir_for(service,
                                  blacklist_paths,
                                  user=_user,
                                  group=_group)
            for name, ext, filename, zipinfo in gen:
                # See if the name should be preprocessed.
                if _preprocess is not None:
                    name = _preprocess(name)
                # construct a name for the output file.
                yaml_filename = path_for_policy_file(service, name)
                if yaml_filename in blacklist_paths:
                    raise BadPolicyZipFile("policy.d name {} is blacklisted"
                                           .format(yaml_filename))
                with zfp.open(zipinfo) as fp:
                    doc = fp.read()
                    # if template_function is not None, then offer the document
                    # to the template function
                    if ext in POLICYD_TEMPLATE_EXTS:
                        if (template_function is None or not
                                callable(template_function)):
                            raise BadPolicyZipFile(
                                "Template {} but no template_function is "
                                "available".format(filename))
                        doc = template_function(doc)
                    yaml_doc = read_and_validate_yaml(doc, blacklist_keys)
                # we may have to create the directory
                maybe_create_directory_for(yaml_filename, _user, _group)
                ch_host.write_file(yaml_filename,
                                   yaml.dump(yaml_doc).encode('utf-8'),
                                   _user,
                                   _group)
        # Every thing worked, so we mark up a success.
        completed = True
    except (zipfile.BadZipFile, BadPolicyZipFile, BadPolicyYamlFile) as e:
        hookenv.log("Processing {} failed: {}".format(resource_file, str(e)),
                    level=POLICYD_LOG_LEVEL_DEFAULT)
    except IOError as e:
        # technically this shouldn't happen; it would be a programming error as
        # the filename comes from Juju and thus, should exist.
        hookenv.log(
            "File {} failed with IOError.  This really shouldn't happen"
            " -- error: {}".format(resource_file, str(e)),
            level=POLICYD_LOG_LEVEL_DEFAULT)
    except Exception as e:
        import traceback
        hookenv.log("General Exception({}) during policyd processing"
                    .format(str(e)),
                    level=POLICYD_LOG_LEVEL_DEFAULT)
        hookenv.log(traceback.format_exc())
    finally:
        if not completed:
            hookenv.log("Processing {} failed: cleaning policy.d directory"
                        .format(resource_file),
                        level=POLICYD_LOG_LEVEL_DEFAULT)
            clean_policyd_dir_for(service,
                                  blacklist_paths,
                                  user=_user,
                                  group=_group)
        else:
            # touch the success filename
            hookenv.log("policy.d overrides installed.",
                        level=POLICYD_LOG_LEVEL_DEFAULT)
            set_policy_success_file()
        return completed



import mock
import types
import uuid

import charmhelpers.contrib.network.ovs as ovs

import tests.utils as test_utils


# NOTE(fnordahl): some functions drectly under the ``contrib.network.ovs``
# module have their unit tests in the ``test_ovs.py`` module in the
# ``tests.contrib.network`` package.


class TestOVS(test_utils.BaseTestCase):

    def test__dict_to_vsctl_set(self):
        indata = {
            'key': 'value',
            'otherkey': {
                'nestedkey': 'nestedvalue',
            },
        }
        # due to varying Dict ordering depending on Python version we need
        # to be a bit elaborate rather than comparing result directly
        result1 = ('--', 'set', 'aTable', 'anItem', 'key=value')
        result2 = ('--', 'set', 'aTable', 'anItem',
                   'otherkey:nestedkey=nestedvalue')
        for setcmd in ovs._dict_to_vsctl_set(indata, 'aTable', 'anItem'):
            self.assertTrue(setcmd == result1 or setcmd == result2)

    def test_add_bridge(self):
        self.patch_object(ovs.subprocess, 'check_call')
        self.patch_object(ovs, 'log')
        ovs.add_bridge('test')
        self.check_call.assert_called_once_with([
            "ovs-vsctl", "--", "--may-exist",
            "add-br", 'test'])
        self.assertTrue(self.log.call_count == 1)

        self.check_call.reset_mock()
        self.log.reset_mock()
        ovs.add_bridge('test', datapath_type='netdev')
        self.check_call.assert_called_with([
            "ovs-vsctl", "--", "--may-exist",
            "add-br", 'test', "--", "set",
            "bridge", "test", "datapath_type=netdev",
        ])
        self.assertTrue(self.log.call_count == 2)

        self.check_call.reset_mock()
        ovs.add_bridge('test', exclusive=True)
        self.check_call.assert_called_once_with([
            "ovs-vsctl", "--", "add-br", 'test'])

        self.check_call.reset_mock()
        self.patch_object(ovs, '_dict_to_vsctl_set')
        self._dict_to_vsctl_set.return_value = [['--', 'fakeextradata']]
        ovs.add_bridge('test', brdata={'fakeinput': None})
        self._dict_to_vsctl_set.assert_called_once_with(
            {'fakeinput': None}, 'bridge', 'test')
        self.check_call.assert_called_once_with([
            'ovs-vsctl', '--', '--may-exist', 'add-br', 'test',
            '--', 'fakeextradata'])

    def test_add_bridge_port(self):
        self.patch_object(ovs.subprocess, 'check_call')
        self.patch_object(ovs, 'log')
        ovs.add_bridge_port('test', 'eth1')
        self.check_call.assert_has_calls([
            mock.call(['ovs-vsctl', '--', '--may-exist', 'add-port',
                       'test', 'eth1']),
            mock.call(['ip', 'link', 'set', 'eth1', 'up']),
            mock.call(['ip', 'link', 'set', 'eth1', 'promisc', 'off'])
        ])
        self.assertTrue(self.log.call_count == 1)

        self.check_call.reset_mock()
        self.log.reset_mock()
        ovs.add_bridge_port('test', 'eth1', promisc=True)
        self.check_call.assert_has_calls([
            mock.call(['ovs-vsctl', '--', '--may-exist', 'add-port',
                       'test', 'eth1']),
            mock.call(['ip', 'link', 'set', 'eth1', 'up']),
            mock.call(['ip', 'link', 'set', 'eth1', 'promisc', 'on'])
        ])
        self.assertTrue(self.log.call_count == 1)

        self.check_call.reset_mock()
        self.log.reset_mock()
        ovs.add_bridge_port('test', 'eth1', promisc=None)
        self.check_call.assert_has_calls([
            mock.call(['ovs-vsctl', '--', '--may-exist', 'add-port',
                       'test', 'eth1']),
            mock.call(['ip', 'link', 'set', 'eth1', 'up']),
        ])
        self.assertTrue(self.log.call_count == 1)

        self.check_call.reset_mock()
        ovs.add_bridge_port('test', 'eth1', exclusive=True, linkup=False)
        self.check_call.assert_has_calls([
            mock.call(['ovs-vsctl', '--', 'add-port', 'test', 'eth1']),
            mock.call(['ip', 'link', 'set', 'eth1', 'promisc', 'off'])
        ])

        self.check_call.reset_mock()
        self.patch_object(ovs, '_dict_to_vsctl_set')
        self._dict_to_vsctl_set.return_value = [['--', 'fakeextradata']]
        ovs.add_bridge_port('test', 'eth1', ifdata={'fakeinput': None})
        self._dict_to_vsctl_set.assert_called_once_with(
            {'fakeinput': None}, 'Interface', 'eth1')
        self.check_call.assert_has_calls([
            mock.call(['ovs-vsctl', '--', '--may-exist', 'add-port',
                       'test', 'eth1', '--', 'fakeextradata']),
            mock.call(['ip', 'link', 'set', 'eth1', 'up']),
            mock.call(['ip', 'link', 'set', 'eth1', 'promisc', 'off'])
        ])
        self._dict_to_vsctl_set.reset_mock()
        self.check_call.reset_mock()
        ovs.add_bridge_port('test', 'eth1', portdata={'fakeportinput': None})
        self._dict_to_vsctl_set.assert_called_once_with(
            {'fakeportinput': None}, 'Port', 'eth1')
        self.check_call.assert_has_calls([
            mock.call(['ovs-vsctl', '--', '--may-exist', 'add-port',
                       'test', 'eth1', '--', 'fakeextradata']),
            mock.call(['ip', 'link', 'set', 'eth1', 'up']),
            mock.call(['ip', 'link', 'set', 'eth1', 'promisc', 'off'])
        ])

    def test_del_bridge_port(self):
        self.patch_object(ovs.subprocess, 'check_call')
        self.patch_object(ovs, 'log')
        ovs.del_bridge_port('test', 'eth1')
        self.check_call.assert_has_calls([
            mock.call(['ovs-vsctl', '--', '--if-exists', 'del-port',
                       'test', 'eth1']),
            mock.call(['ip', 'link', 'set', 'eth1', 'down']),
            mock.call(['ip', 'link', 'set', 'eth1', 'promisc', 'off'])
        ])
        self.assertTrue(self.log.call_count == 1)
        self.assertTrue(self.check_call.call_count == 3)
        self.check_call.reset_mock()
        ovs.del_bridge_port('test', 'eth1', linkdown=False)
        self.check_call.assert_called_once_with(
            ['ovs-vsctl', '--', '--if-exists', 'del-port', 'test', 'eth1'])

    def test_ovs_appctl(self):
        self.patch_object(ovs.subprocess, 'check_output')
        ovs.ovs_appctl('ovs-vswitchd', ('ofproto/list',))
        self.check_output.assert_called_once_with(
            ['ovs-appctl', '-t', 'ovs-vswitchd', 'ofproto/list'],
            universal_newlines=True)

    def test_add_bridge_bond(self):
        self.patch_object(ovs.subprocess, 'check_call')
        self.patch_object(ovs, '_dict_to_vsctl_set')
        self._dict_to_vsctl_set.return_value = [['--', 'fakekey=fakevalue']]
        portdata = {
            'bond-mode': 'balance-tcp',
            'lacp': 'active',
            'other-config': {
                'lacp-time': 'fast',
            },
        }
        ifdatamap = {
            'eth0': {
                'type': 'dpdk',
                'mtu-request': '9000',
                'options': {
                    'dpdk-devargs': '0000:01:00.0',
                },
            },
            'eth1': {
                'type': 'dpdk',
                'mtu-request': '9000',
                'options': {
                    'dpdk-devargs': '0000:02:00.0',
                },
            },
        }
        ovs.add_bridge_bond('br-ex', 'bond42', ['eth0', 'eth1'],
                            portdata, ifdatamap)
        self._dict_to_vsctl_set.assert_has_calls([
            mock.call(portdata, 'port', 'bond42'),
            mock.call(ifdatamap['eth0'], 'Interface', 'eth0'),
            mock.call(ifdatamap['eth1'], 'Interface', 'eth1'),
        ], any_order=True)
        self.check_call.assert_called_once_with([
            'ovs-vsctl',
            '--', '--may-exist', 'add-bond', 'br-ex', 'bond42', 'eth0', 'eth1',
            '--', 'fakekey=fakevalue',
            '--', 'fakekey=fakevalue',
            '--', 'fakekey=fakevalue'])

    def test_uuid_for_port(self):
        self.patch_object(ovs.ch_ovsdb, 'SimpleOVSDB')
        fake_uuid = uuid.UUID('efdce2cf-cd66-4060-a9f8-1db0e9a06216')
        ovsdb = mock.MagicMock()
        ovsdb.port.find.return_value = [
            {'_uuid': fake_uuid},
        ]
        self.SimpleOVSDB.return_value = ovsdb
        self.assertEquals(ovs.uuid_for_port('fake-port'), fake_uuid)
        ovsdb.port.find.assert_called_once_with('name=fake-port')

    def test_bridge_for_port(self):
        self.patch_object(ovs.ch_ovsdb, 'SimpleOVSDB')
        fake_uuid = uuid.UUID('818d03dd-efb8-44be-aba3-bde423bf1cc9')
        ovsdb = mock.MagicMock()
        ovsdb.bridge.__iter__.return_value = [
            {
                'name': 'fake-bridge',
                'ports': [fake_uuid],
            },
        ]
        self.SimpleOVSDB.return_value = ovsdb
        self.assertEquals(ovs.bridge_for_port(fake_uuid), 'fake-bridge')
        # If there is a single port on a bridge the ports property will not be
        # a list. ref: juju/charm-helpers#510
        ovsdb.bridge.__iter__.return_value = [
            {
                'name': 'fake-bridge',
                'ports': fake_uuid,
            },
        ]
        self.assertEquals(ovs.bridge_for_port(fake_uuid), 'fake-bridge')

    def test_patch_ports_on_bridge(self):
        self.patch_object(ovs.ch_ovsdb, 'SimpleOVSDB')
        self.patch_object(ovs, 'bridge_for_port')
        self.patch_object(ovs, 'uuid_for_port')
        ovsdb = mock.MagicMock()
        ovsdb.interface.find.return_value = [
            {
                'name': 'fake-interface-with-port-for-other-bridge',
                'options': {
                    'peer': 'fake-peer'
                },
            },
            {
                'name': 'fake-interface',
                'options': {
                    'peer': 'fake-peer'
                },
            },
        ]
        port_uuid = uuid.UUID('0d43905b-f80e-4eaa-9feb-a9017da8c6bc')
        ovsdb.port.find.side_effect = [
            [{
                '_uuid': port_uuid,
                'name': 'port-on-other-bridge',
            }],
            [{
                '_uuid': port_uuid,
                'name': 'fake-port',
            }],
        ]
        self.SimpleOVSDB.return_value = ovsdb
        self.bridge_for_port.side_effect = ['some-other-bridge', 'fake-bridge', 'fake-peer-bridge']
        for patch in ovs.patch_ports_on_bridge('fake-bridge'):
            self.assertEquals(
                patch,
                ovs.Patch(
                    this_end=ovs.PatchPort(
                        bridge='fake-bridge',
                        port='fake-port'),
                    other_end=ovs.PatchPort(
                        bridge='fake-peer-bridge',
                        port='fake-peer'))
            )
            break
        else:
            assert 0, 'Expected generator to provide output'
        ovsdb.port.find.side_effect = None
        ovsdb.port.find.return_value = []
        with self.assertRaises(ValueError):
            for patch in ovs.patch_ports_on_bridge('fake-bridge'):
                pass
        ovsdb.interface.find.return_value = []
        for patch in ovs.patch_ports_on_bridge('fake-bridge'):
            assert 0, 'Expected generator to provide empty iterator'
        self.assertTrue(isinstance(
            ovs.patch_ports_on_bridge('fake-bridge'), types.GeneratorType))



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Copyright 2014-2021 Canonical Limited.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#  http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# Authors:
#  Kapil Thangavelu <kapil.foss@gmail.com>
#
"""
Intro
-----

A simple way to store state in units. This provides a key value
storage with support for versioned, transactional operation,
and can calculate deltas from previous values to simplify unit logic
when processing changes.


Hook Integration
----------------

There are several extant frameworks for hook execution, including

 - charmhelpers.core.hookenv.Hooks
 - charmhelpers.core.services.ServiceManager

The storage classes are framework agnostic, one simple integration is
via the HookData contextmanager. It will record the current hook
execution environment (including relation data, config data, etc.),
setup a transaction and allow easy access to the changes from
previously seen values. One consequence of the integration is the
reservation of particular keys ('rels', 'unit', 'env', 'config',
'charm_revisions') for their respective values.

Here's a fully worked integration example using hookenv.Hooks::

       from charmhelper.core import hookenv, unitdata

       hook_data = unitdata.HookData()
       db = unitdata.kv()
       hooks = hookenv.Hooks()

       @hooks.hook
       def config_changed():
           # Print all changes to configuration from previously seen
           # values.
           for changed, (prev, cur) in hook_data.conf.items():
               print('config changed', changed,
                     'previous value', prev,
                     'current value',  cur)

           # Get some unit specific bookkeeping
           if not db.get('pkg_key'):
               key = urllib.urlopen('https://example.com/pkg_key').read()
               db.set('pkg_key', key)

           # Directly access all charm config as a mapping.
           conf = db.getrange('config', True)

           # Directly access all relation data as a mapping
           rels = db.getrange('rels', True)

       if __name__ == '__main__':
           with hook_data():
               hook.execute()


A more basic integration is via the hook_scope context manager which simply
manages transaction scope (and records hook name, and timestamp)::

  >>> from unitdata import kv
  >>> db = kv()
  >>> with db.hook_scope('install'):
  ...    # do work, in transactional scope.
  ...    db.set('x', 1)
  >>> db.get('x')
  1


Usage
-----

Values are automatically json de/serialized to preserve basic typing
and complex data struct capabilities (dicts, lists, ints, booleans, etc).

Individual values can be manipulated via get/set::

   >>> kv.set('y', True)
   >>> kv.get('y')
   True

   # We can set complex values (dicts, lists) as a single key.
   >>> kv.set('config', {'a': 1, 'b': True'})

   # Also supports returning dictionaries as a record which
   # provides attribute access.
   >>> config = kv.get('config', record=True)
   >>> config.b
   True


Groups of keys can be manipulated with update/getrange::

   >>> kv.update({'z': 1, 'y': 2}, prefix="gui.")
   >>> kv.getrange('gui.', strip=True)
   {'z': 1, 'y': 2}

When updating values, its very helpful to understand which values
have actually changed and how have they changed. The storage
provides a delta method to provide for this::

   >>> data = {'debug': True, 'option': 2}
   >>> delta = kv.delta(data, 'config.')
   >>> delta.debug.previous
   None
   >>> delta.debug.current
   True
   >>> delta
   {'debug': (None, True), 'option': (None, 2)}

Note the delta method does not persist the actual change, it needs to
be explicitly saved via 'update' method::

   >>> kv.update(data, 'config.')

Values modified in the context of a hook scope retain historical values
associated to the hookname.

   >>> with db.hook_scope('config-changed'):
   ...      db.set('x', 42)
   >>> db.gethistory('x')
   [(1, u'x', 1, u'install', u'2015-01-21T16:49:30.038372'),
    (2, u'x', 42, u'config-changed', u'2015-01-21T16:49:30.038786')]

"""

import collections
import contextlib
import datetime
import itertools
import json
import os
import pprint
import sqlite3
import sys

__author__ = 'Kapil Thangavelu <kapil.foss@gmail.com>'


class Storage(object):
    """Simple key value database for local unit state within charms.

    Modifications are not persisted unless :meth:`flush` is called.

    To support dicts, lists, integer, floats, and booleans values
    are automatically json encoded/decoded.

    Note: to facilitate unit testing, ':memory:' can be passed as the
    path parameter which causes sqlite3 to only build the db in memory.
    This should only be used for testing purposes.
    """
    def __init__(self, path=None):
        self.db_path = path
        if path is None:
            if 'UNIT_STATE_DB' in os.environ:
                self.db_path = os.environ['UNIT_STATE_DB']
            else:
                self.db_path = os.path.join(
                    os.environ.get('CHARM_DIR', ''), '.unit-state.db')
        if self.db_path != ':memory:':
            with open(self.db_path, 'a') as f:
                os.fchmod(f.fileno(), 0o600)
        self.conn = sqlite3.connect('%s' % self.db_path)
        self.cursor = self.conn.cursor()
        self.revision = None
        self._closed = False
        self._init()

    def close(self):
        if self._closed:
            return
        self.flush(False)
        self.cursor.close()
        self.conn.close()
        self._closed = True

    def get(self, key, default=None, record=False):
        self.cursor.execute('select data from kv where key=?', [key])
        result = self.cursor.fetchone()
        if not result:
            return default
        if record:
            return Record(json.loads(result[0]))
        return json.loads(result[0])

    def getrange(self, key_prefix, strip=False):
        """
        Get a range of keys starting with a common prefix as a mapping of
        keys to values.

        :param str key_prefix: Common prefix among all keys
        :param bool strip: Optionally strip the common prefix from the key
            names in the returned dict
        :return dict: A (possibly empty) dict of key-value mappings
        """
        self.cursor.execute("select key, data from kv where key like ?",
                            ['%s%%' % key_prefix])
        result = self.cursor.fetchall()

        if not result:
            return {}
        if not strip:
            key_prefix = ''
        return dict([
            (k[len(key_prefix):], json.loads(v)) for k, v in result])

    def update(self, mapping, prefix=""):
        """
        Set the values of multiple keys at once.

        :param dict mapping: Mapping of keys to values
        :param str prefix: Optional prefix to apply to all keys in `mapping`
            before setting
        """
        for k, v in mapping.items():
            self.set("%s%s" % (prefix, k), v)

    def unset(self, key):
        """
        Remove a key from the database entirely.
        """
        self.cursor.execute('delete from kv where key=?', [key])
        if self.revision and self.cursor.rowcount:
            self.cursor.execute(
                'insert into kv_revisions values (?, ?, ?)',
                [key, self.revision, json.dumps('DELETED')])

    def unsetrange(self, keys=None, prefix=""):
        """
        Remove a range of keys starting with a common prefix, from the database
        entirely.

        :param list keys: List of keys to remove.
        :param str prefix: Optional prefix to apply to all keys in ``keys``
            before removing.
        """
        if keys is not None:
            keys = ['%s%s' % (prefix, key) for key in keys]
            self.cursor.execute('delete from kv where key in (%s)' % ','.join(['?'] * len(keys)), keys)
            if self.revision and self.cursor.rowcount:
                self.cursor.execute(
                    'insert into kv_revisions values %s' % ','.join(['(?, ?, ?)'] * len(keys)),
                    list(itertools.chain.from_iterable((key, self.revision, json.dumps('DELETED')) for key in keys)))
        else:
            self.cursor.execute('delete from kv where key like ?',
                                ['%s%%' % prefix])
            if self.revision and self.cursor.rowcount:
                self.cursor.execute(
                    'insert into kv_revisions values (?, ?, ?)',
                    ['%s%%' % prefix, self.revision, json.dumps('DELETED')])

    def set(self, key, value):
        """
        Set a value in the database.

        :param str key: Key to set the value for
        :param value: Any JSON-serializable value to be set
        """
        serialized = json.dumps(value)

        self.cursor.execute('select data from kv where key=?', [key])
        exists = self.cursor.fetchone()

        # Skip mutations to the same value
        if exists:
            if exists[0] == serialized:
                return value

        if not exists:
            self.cursor.execute(
                'insert into kv (key, data) values (?, ?)',
                (key, serialized))
        else:
            self.cursor.execute('''
            update kv
            set data = ?
            where key = ?''', [serialized, key])

        # Save
        if not self.revision:
            return value

        self.cursor.execute(
            'select 1 from kv_revisions where key=? and revision=?',
            [key, self.revision])
        exists = self.cursor.fetchone()

        if not exists:
            self.cursor.execute(
                '''insert into kv_revisions (
                revision, key, data) values (?, ?, ?)''',
                (self.revision, key, serialized))
        else:
            self.cursor.execute(
                '''
                update kv_revisions
                set data = ?
                where key = ?
                and   revision = ?''',
                [serialized, key, self.revision])

        return value

    def delta(self, mapping, prefix):
        """
        return a delta containing values that have changed.
        """
        previous = self.getrange(prefix, strip=True)
        if not previous:
            pk = set()
        else:
            pk = set(previous.keys())
        ck = set(mapping.keys())
        delta = DeltaSet()

        # added
        for k in ck.difference(pk):
            delta[k] = Delta(None, mapping[k])

        # removed
        for k in pk.difference(ck):
            delta[k] = Delta(previous[k], None)

        # changed
        for k in pk.intersection(ck):
            c = mapping[k]
            p = previous[k]
            if c != p:
                delta[k] = Delta(p, c)

        return delta

    @contextlib.contextmanager
    def hook_scope(self, name=""):
        """Scope all future interactions to the current hook execution
        revision."""
        assert not self.revision
        self.cursor.execute(
            'insert into hooks (hook, date) values (?, ?)',
            (name or sys.argv[0],
             datetime.datetime.utcnow().isoformat()))
        self.revision = self.cursor.lastrowid
        try:
            yield self.revision
            self.revision = None
        except Exception:
            self.flush(False)
            self.revision = None
            raise
        else:
            self.flush()

    def flush(self, save=True):
        if save:
            self.conn.commit()
        elif self._closed:
            return
        else:
            self.conn.rollback()

    def _init(self):
        self.cursor.execute('''
            create table if not exists kv (
               key text,
               data text,
               primary key (key)
               )''')
        self.cursor.execute('''
            create table if not exists kv_revisions (
               key text,
               revision integer,
               data text,
               primary key (key, revision)
               )''')
        self.cursor.execute('''
            create table if not exists hooks (
               version integer primary key autoincrement,
               hook text,
               date text
               )''')
        self.conn.commit()

    def gethistory(self, key, deserialize=False):
        self.cursor.execute(
            '''
            select kv.revision, kv.key, kv.data, h.hook, h.date
            from kv_revisions kv,
                 hooks h
            where kv.key=?
             and kv.revision = h.version
            ''', [key])
        if deserialize is False:
            return self.cursor.fetchall()
        return map(_parse_history, self.cursor.fetchall())

    def debug(self, fh=sys.stderr):
        self.cursor.execute('select * from kv')
        pprint.pprint(self.cursor.fetchall(), stream=fh)
        self.cursor.execute('select * from kv_revisions')
        pprint.pprint(self.cursor.fetchall(), stream=fh)


def _parse_history(d):
    return (d[0], d[1], json.loads(d[2]), d[3],
            datetime.datetime.strptime(d[-1], "%Y-%m-%dT%H:%M:%S.%f"))


class HookData(object):
    """Simple integration for existing hook exec frameworks.

    Records all unit information, and stores deltas for processing
    by the hook.

    Sample::

       from charmhelper.core import hookenv, unitdata

       changes = unitdata.HookData()
       db = unitdata.kv()
       hooks = hookenv.Hooks()

       @hooks.hook
       def config_changed():
           # View all changes to configuration
           for changed, (prev, cur) in changes.conf.items():
               print('config changed', changed,
                     'previous value', prev,
                     'current value',  cur)

           # Get some unit specific bookkeeping
           if not db.get('pkg_key'):
               key = urllib.urlopen('https://example.com/pkg_key').read()
               db.set('pkg_key', key)

       if __name__ == '__main__':
           with changes():
               hook.execute()

    """
    def __init__(self):
        self.kv = kv()
        self.conf = None
        self.rels = None

    @contextlib.contextmanager
    def __call__(self):
        from charmhelpers.core import hookenv
        hook_name = hookenv.hook_name()

        with self.kv.hook_scope(hook_name):
            self._record_charm_version(hookenv.charm_dir())
            delta_config, delta_relation = self._record_hook(hookenv)
            yield self.kv, delta_config, delta_relation

    def _record_charm_version(self, charm_dir):
        # Record revisions.. charm revisions are meaningless
        # to charm authors as they don't control the revision.
        # so logic dependnent on revision is not particularly
        # useful, however it is useful for debugging analysis.
        charm_rev = open(
            os.path.join(charm_dir, 'revision')).read().strip()
        charm_rev = charm_rev or '0'
        revs = self.kv.get('charm_revisions', [])
        if charm_rev not in revs:
            revs.append(charm_rev.strip() or '0')
            self.kv.set('charm_revisions', revs)

    def _record_hook(self, hookenv):
        data = hookenv.execution_environment()
        self.conf = conf_delta = self.kv.delta(data['conf'], 'config')
        self.rels = rels_delta = self.kv.delta(data['rels'], 'rels')
        self.kv.set('env', dict(data['env']))
        self.kv.set('unit', data['unit'])
        self.kv.set('relid', data.get('relid'))
        return conf_delta, rels_delta


class Record(dict):

    __slots__ = ()

    def __getattr__(self, k):
        if k in self:
            return self[k]
        raise AttributeError(k)


class DeltaSet(Record):

    __slots__ = ()


Delta = collections.namedtuple('Delta', ['previous', 'current'])


_KV = None


def kv():
    global _KV
    if _KV is None:
        _KV = Storage()
    return _KV



import tempfile
import copy
import mock
import unittest
import shutil
import yaml

from charmhelpers.contrib.openstack import policy_rcd

TEST_POLICY = {
    'blocked_actions': {
        'neutron-dhcp-agent': ['restart', 'stop', 'try-restart'],
        'neutron-l3-agent': ['restart', 'stop', 'try-restart'],
        'neutron-metadata-agent': ['restart', 'stop', 'try-restart'],
        'neutron-openvswitch-agent': ['restart', 'stop', 'try-restart'],
        'openvswitch-switch': ['restart', 'stop', 'try-restart'],
        'ovs-vswitchd': ['restart', 'stop', 'try-restart'],
        'ovs-vswitchd-dpdk': ['restart', 'stop', 'try-restart'],
        'ovsdb-server': ['restart', 'stop', 'try-restart']},
    'policy_requestor_name': 'neutron-openvswitch',
    'policy_requestor_type': 'charm'}


class PolicyRCDTests(unittest.TestCase):

    def setUp(self):
        super(PolicyRCDTests, self).setUp()
        self.tmp_dir = tempfile.mkdtemp()
        self.addCleanup(lambda: shutil.rmtree(self.tmp_dir))

    @mock.patch.object(policy_rcd.hookenv, "service_name")
    def test_get_policy_file_name(self, service_name):
        service_name.return_value = 'mysvc'
        self.assertEqual(
            policy_rcd.get_policy_file_name(),
            '/etc/policy-rc.d/charm-mysvc.policy')

    @mock.patch.object(policy_rcd, "get_policy_file_name")
    def test_read_default_policy_file(self, get_policy_file_name):
        with tempfile.NamedTemporaryFile('w') as ftmp:
            policy_rcd.write_policy_file(ftmp.name, TEST_POLICY)
            get_policy_file_name.return_value = ftmp.name
            self.assertEqual(
                policy_rcd.read_default_policy_file(),
                TEST_POLICY)

    def test_write_policy_file(self):
        with tempfile.NamedTemporaryFile('w') as ftmp:
            policy_rcd.write_policy_file(ftmp.name, TEST_POLICY)
            with open(ftmp.name, 'r') as f:
                policy = yaml.safe_load(f)
            self.assertEqual(policy, TEST_POLICY)

    @mock.patch.object(policy_rcd.os, "remove")
    @mock.patch.object(policy_rcd.hookenv, "service_name")
    def test_remove_policy_file(self, service_name, remove):
        service_name.return_value = 'mysvc'
        policy_rcd.remove_policy_file()
        remove.assert_called_once_with('/etc/policy-rc.d/charm-mysvc.policy')

    @mock.patch.object(policy_rcd.os.path, "exists")
    @mock.patch.object(policy_rcd.shutil, "copy2")
    @mock.patch.object(policy_rcd.host, "mkdir")
    @mock.patch.object(policy_rcd.alternatives, "install_alternative")
    @mock.patch.object(policy_rcd.hookenv, "service_name")
    @mock.patch.object(policy_rcd.os.path, "dirname")
    def test_install_policy_rcd(self, dirname, service_name,
                                install_alternative, mkdir, copy2, exists):
        dirs = ['/dir1', '/dir2']
        service_name.return_value = 'mysvc'
        dirname.side_effect = lambda x: dirs.pop()
        exists.return_value = False
        policy_rcd.install_policy_rcd()
        install_alternative.assert_called_once_with(
            'policy-rc.d',
            '/usr/sbin/policy-rc.d',
            '/var/lib/charm/mysvc/policy-rc.d')
        mkdir.assert_has_calls([
            mock.call('/dir1'),
            mock.call('/etc/policy-rc.d')
        ])
        copy2.assert_called_once_with(
            '/dir2/policy_rc_d_script.py',
            '/var/lib/charm/mysvc/policy-rc.d')

    @mock.patch.object(policy_rcd.hookenv, "service_name")
    def test_get_default_policy(self, service_name):
        service_name.return_value = 'mysvc'
        self.assertEqual(
            policy_rcd.get_default_policy(),
            {
                'policy_requestor_name': 'mysvc',
                'policy_requestor_type': 'charm',
                'blocked_actions': {}})

    @mock.patch.object(policy_rcd, "write_policy_file")
    @mock.patch.object(policy_rcd.hookenv, "service_name")
    @mock.patch.object(policy_rcd, "read_default_policy_file")
    def test_add_policy_block(self, read_default_policy_file, service_name,
                              write_policy_file):
        service_name.return_value = 'mysvc'
        old_policy = copy.deepcopy(TEST_POLICY)
        read_default_policy_file.return_value = old_policy
        policy_rcd.add_policy_block('apache2', ['restart'])
        expect_policy = copy.deepcopy(TEST_POLICY)
        expect_policy['blocked_actions']['apache2'] = ['restart']
        write_policy_file.assert_called_once_with(
            '/etc/policy-rc.d/charm-mysvc.policy',
            expect_policy)

    @mock.patch.object(policy_rcd, "write_policy_file")
    @mock.patch.object(policy_rcd.hookenv, "service_name")
    @mock.patch.object(policy_rcd, "read_default_policy_file")
    def test_remove_policy_block(self, read_default_policy_file, service_name,
                                 write_policy_file):
        service_name.return_value = 'mysvc'
        old_policy = copy.deepcopy(TEST_POLICY)
        read_default_policy_file.return_value = old_policy
        policy_rcd.remove_policy_block(
            'neutron-dhcp-agent',
            ['try-restart', 'restart'])
        expect_policy = copy.deepcopy(TEST_POLICY)
        expect_policy['blocked_actions']['neutron-dhcp-agent'] = ['stop']
        write_policy_file.assert_called_once_with(
            '/etc/policy-rc.d/charm-mysvc.policy',
            expect_policy)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

from charmhelpers.core import files

import mock
import unittest
import tempfile
import os


class FileTests(unittest.TestCase):

    @mock.patch("subprocess.check_call")
    def test_sed(self, check_call):
        files.sed("/tmp/test-sed-file", "replace", "this")
        check_call.assert_called_once_with(
            ['sed', '-i', '-r', '-e', 's/replace/this/g',
             '/tmp/test-sed-file']
        )

    def test_sed_file(self):
        tmp = tempfile.NamedTemporaryFile(mode='w', delete=False)
        tmp.write("IPV6=yes")
        tmp.close()

        files.sed(tmp.name, "IPV6=.*", "IPV6=no")

        with open(tmp.name) as tmp:
            self.assertEquals(tmp.read(), "IPV6=no")

        os.unlink(tmp.name)



import subprocess
import yum
import os

from charmhelpers.core.strutils import BasicStringComparator


class CompareHostReleases(BasicStringComparator):
    """Provide comparisons of Host releases.

    Use in the form of

    if CompareHostReleases(release) > 'trusty':
        # do something with mitaka
    """

    def __init__(self, item):
        raise NotImplementedError(
            "CompareHostReleases() is not implemented for CentOS")


def service_available(service_name):
    # """Determine whether a system service is available."""
    if os.path.isdir('/run/systemd/system'):
        cmd = ['systemctl', 'is-enabled', service_name]
    else:
        cmd = ['service', service_name, 'is-enabled']
    return subprocess.call(cmd) == 0


def add_new_group(group_name, system_group=False, gid=None):
    cmd = ['groupadd']
    if gid:
        cmd.extend(['--gid', str(gid)])
    if system_group:
        cmd.append('-r')
    cmd.append(group_name)
    subprocess.check_call(cmd)


def lsb_release():
    """Return /etc/os-release in a dict."""
    d = {}
    with open('/etc/os-release', 'r') as lsb:
        for l in lsb:
            s = l.split('=')
            if len(s) != 2:
                continue
            d[s[0].strip()] = s[1].strip()
    return d


def cmp_pkgrevno(package, revno, pkgcache=None):
    """Compare supplied revno with the revno of the installed package.

    *  1 => Installed revno is greater than supplied arg
    *  0 => Installed revno is the same as supplied arg
    * -1 => Installed revno is less than supplied arg

    This function imports YumBase function if the pkgcache argument
    is None.
    """
    if not pkgcache:
        y = yum.YumBase()
        packages = y.doPackageLists()
        pkgcache = {i.Name: i.version for i in packages['installed']}
    pkg = pkgcache[package]
    if pkg > revno:
        return 1
    if pkg < revno:
        return -1
    return 0



# Copyright 2014-2021 Canonical Limited.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#  http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import functools
from subprocess import (
    CalledProcessError,
    check_call,
    check_output,
    Popen,
    PIPE,
)


##################################################
# LVM helpers.
##################################################
def deactivate_lvm_volume_group(block_device):
    '''
    Deactivate any volume group associated with an LVM physical volume.

    :param block_device: str: Full path to LVM physical volume
    '''
    vg = list_lvm_volume_group(block_device)
    if vg:
        cmd = ['vgchange', '-an', vg]
        check_call(cmd)


def is_lvm_physical_volume(block_device):
    '''
    Determine whether a block device is initialized as an LVM PV.

    :param block_device: str: Full path of block device to inspect.

    :returns: boolean: True if block device is a PV, False if not.
    '''
    try:
        check_output(['pvdisplay', block_device])
        return True
    except CalledProcessError:
        return False


def remove_lvm_physical_volume(block_device):
    '''
    Remove LVM PV signatures from a given block device.

    :param block_device: str: Full path of block device to scrub.
    '''
    p = Popen(['pvremove', '-ff', block_device],
              stdin=PIPE)
    p.communicate(input='y\n')


def list_lvm_volume_group(block_device):
    '''
    List LVM volume group associated with a given block device.

    Assumes block device is a valid LVM PV.

    :param block_device: str: Full path of block device to inspect.

    :returns: str: Name of volume group associated with block device or None
    '''
    vg = None
    pvd = check_output(['pvdisplay', block_device]).splitlines()
    for lvm in pvd:
        lvm = lvm.decode('UTF-8')
        if lvm.strip().startswith('VG Name'):
            vg = ' '.join(lvm.strip().split()[2:])
    return vg


def create_lvm_physical_volume(block_device):
    '''
    Initialize a block device as an LVM physical volume.

    :param block_device: str: Full path of block device to initialize.

    '''
    check_call(['pvcreate', block_device])


def create_lvm_volume_group(volume_group, block_device):
    '''
    Create an LVM volume group backed by a given block device.

    Assumes block device has already been initialized as an LVM PV.

    :param volume_group: str: Name of volume group to create.
    :block_device: str: Full path of PV-initialized block device.
    '''
    check_call(['vgcreate', volume_group, block_device])


def list_logical_volumes(select_criteria=None, path_mode=False):
    '''
    List logical volumes

    :param select_criteria: str: Limit list to those volumes matching this
                                 criteria (see 'lvs -S help' for more details)
    :param path_mode: bool: return logical volume name in 'vg/lv' format, this
                            format is required for some commands like lvextend
    :returns: [str]: List of logical volumes
    '''
    lv_diplay_attr = 'lv_name'
    if path_mode:
        # Parsing output logic relies on the column order
        lv_diplay_attr = 'vg_name,' + lv_diplay_attr
    cmd = ['lvs', '--options', lv_diplay_attr, '--noheadings']
    if select_criteria:
        cmd.extend(['--select', select_criteria])
    lvs = []
    for lv in check_output(cmd).decode('UTF-8').splitlines():
        if not lv:
            continue
        if path_mode:
            lvs.append('/'.join(lv.strip().split()))
        else:
            lvs.append(lv.strip())
    return lvs


list_thin_logical_volume_pools = functools.partial(
    list_logical_volumes,
    select_criteria='lv_attr =~ ^t')

list_thin_logical_volumes = functools.partial(
    list_logical_volumes,
    select_criteria='lv_attr =~ ^V')


def extend_logical_volume_by_device(lv_name, block_device):
    '''
    Extends the size of logical volume lv_name by the amount of free space on
    physical volume block_device.

    :param lv_name: str: name of logical volume to be extended (vg/lv format)
    :param block_device: str: name of block_device to be allocated to lv_name
    '''
    cmd = ['lvextend', lv_name, block_device]
    check_call(cmd)


def create_logical_volume(lv_name, volume_group, size=None):
    '''
    Create a new logical volume in an existing volume group

    :param lv_name: str: name of logical volume to be created.
    :param volume_group: str: Name of volume group to use for the new volume.
    :param size: str: Size of logical volume to create (100% if not supplied)
    :raises subprocess.CalledProcessError: in the event that the lvcreate fails.
    '''
    if size:
        check_call([
            'lvcreate',
            '--yes',
            '-L',
            '{}'.format(size),
            '-n', lv_name, volume_group
        ])
    # create the lv with all the space available, this is needed because the
    # system call is different for LVM
    else:
        check_call([
            'lvcreate',
            '--yes',
            '-l',
            '100%FREE',
            '-n', lv_name, volume_group
        ])
